import hal

from wpilib.iterativerobot import IterativeRobot
from wpilib.command.scheduler import Scheduler
from wpilib.livewindow import LiveWindow


class CommandBasedRobot(IterativeRobot):
    '''
    The base class for a Command-Based Robot. To use, instantiate commands and
    trigger them.
    '''

    def startCompetition(self):
        """
        Provide an alternate "main loop" via startCompetition(). Rewritten
        from IterativeRobot for readability and to initialize scheduler.
        """

        hal.report(hal.HALUsageReporting.kResourceType_Framework,
                   hal.HALUsageReporting.kFramework_Iterative)

        self.scheduler = Scheduler.getInstance()
        self.robotInit()

        # Tell the DS that the robot is ready to be enabled
        hal.observeUserProgramStarting()

        # loop forever, calling the appropriate mode-dependent function
        while True:
            if self.ds.isDisabled():
                hal.observeUserProgramDisabled()
                self.disabledInit()
                while self.ds.isDisabled():
                    self.disabledPeriodic()
                    self.ds.waitForData()

            elif self.ds.isAutonomous():
                hal.observeUserProgramAutonomous()
                self.autonomousInit()
                while self.ds.isEnabled() and self.ds.isAutonomous():
                    self.autonomousPeriodic()
                    self.ds.waitForData()

            elif self.ds.isTest():
                hal.observeUserProgramTest()
                LiveWindow.setEnabled(True)

                self.testInit()
                while self.ds.isEnabled() and self.ds.isTest():
                    self.testPeriodic()
                    self.ds.waitForData()

                LiveWindow.setEnabled(False)

            else:
                hal.observeUserProgramTeleop()
                self.teleopInit()
                # isOperatorControl checks "not autonomous or test", so we need
                # to check isEnabled as well, since otherwise it will continue
                # looping while disabled.
                while self.ds.isEnabled() and self.ds.isOperatorControl():
                    self.teleopPeriodic()
                    self.ds.waitForData()


    def commandPeriodic(self):
        '''
        Run the scheduler regularly. If an error occurs during a competition,
        prevent it from crashing the program.
        '''

        try:
            self.scheduler.run()
        except Exception as error:
            if not self.ds.isFMSAttached():
                raise

            '''Just to be safe, stop all running commands.'''
            self.scheduler.removeAll()

            self.handleCrash(error)


    autonomousPeriodic = commandPeriodic
    teleopPeriodic = commandPeriodic
    disabledPeriodic = commandPeriodic


    def testPeriodic(self):
        '''
        Test mode will not run normal commands, but motors can be controlled
        and sensors viewed with the SmartDashboard.
        '''

        LiveWindow.run()


    def handleCrash(self, error):
        '''
        Called if an exception is raised in the Scheduler during a competition.
        Writes an error message to the driver station by default. If you want
        more complex behavior, override this method in your robot class.
        '''

        self.ds.reportError(str(error), printTrace=True)



import wpilib

class ButtonDebouncer:
    '''Useful utility class for debouncing buttons'''
    
    def __init__(self, joystick, buttonnum, period=0.5):
        '''
        :param joystick:  Joystick object
        :type  joystick:  :class:`wpilib.Joystick` 
        :param buttonnum: Number of button to retrieve
        :type  buttonnum: int
        :param period:    Period of time (in seconds) to wait before allowing new button
                          presses. Defaults to 0.5 seconds.
        :type  period:    float
        '''
        self.joystick = joystick
        self.buttonnum = buttonnum
        self.latest = 0
        self.debounce_period = float(period)
        self.timer = wpilib.Timer
    
    def set_debounce_period(self, period):
        '''Set number of seconds to wait before returning True for the
        button again'''
        self.debounce_period = float(period)
    
    def get(self):
        '''Returns the value of the joystick button. If the button is held down, then
        True will only be returned once every ``debounce_period`` seconds'''
        
        now = self.timer.getFPGATimestamp()
        if self.joystick.getRawButton(self.buttonnum):
            if (now-self.latest) > self.debounce_period: 
                self.latest = now
                return True
        return False



from networktables import NetworkTables

# Only used as a marker
class _TunableProperty(property):
    pass

class _AutosendProperty(_TunableProperty):
    pass

def tunable(default, *, writeDefault=True, subtable=None):
    '''
        This allows you to define simple properties that allow you to easily
        communicate with other programs via NetworkTables.
    
        The following example will define a NetworkTable variable at
        ``/components/my_component/foo``::
        
            class MyRobot(magicbot.MagicRobot):
            
                my_component = MyComponent
        
            ... 
            
            from magicbot import tunable
        
            class MyComponent:
        
                # define the tunable property
                foo = tunable(True)
    
    
                def execute(self):
                
                    # set the variable
                    self.foo = True
                    
                    # get the variable
                    foo = self.foo

        The key of the NetworkTables variable will vary based on what kind of
        object the decorated method belongs to:

        * A component: ``/components/COMPONENTNAME/VARNAME``
        * An autonomous mode: ``/autonomous/MODENAME/VARNAME``
        * Your main robot class: ``/robot/VARNAME``
                    
        .. note:: When executing unit tests on objects that create tunables,
                  you will want to use setup_tunables to set the object up.
                  In normal usage, MagicRobot does this for you, so you don't
                  have to do anything special.
    '''
    
    # the way this works is we use a special class to indicate that it
    # is a tunable, and MagicRobot adds _ntattr and _global_table variables
    # to the class property
    
    # The tricky bit is that you need to do late binding on these, because
    # the networktables key is not known when the object is created. Instead,
    # the name of the key is related to the name of the variable name in the
    # robot class
    
    nt = NetworkTables
    
    def _get(self):
        return getattr(self, prop._ntattr).value
    
    def _set(self, value):
        v = getattr(self, prop._ntattr)
        nt._api.setEntryValue(v.key, v._valuefn(value))
        
    prop = _TunableProperty(fget=_get, fset=_set)
    prop._ntdefault = default
    prop._ntsubtable = subtable
    prop._ntwritedefault = writeDefault
    
    return prop 


def setup_tunables(component, cname, prefix='components'):
    '''
        Connects the tunables on an object to NetworkTables.

        :param component:   Component object
        :param cname:       Name of component
        :type cname: str
        :param prefix:      Prefix to use, or no prefix if None
        :type prefix: str    
    
        .. note:: This is not needed in normal use, only useful
                  for testing
    '''
    
    cls = component.__class__

    if prefix is None:
        prefix = '/%s' % cname
    else:
        prefix = '/%s/%s' % (prefix, cname)
    
    for n in dir(cls):
        if n.startswith('_'):
            continue
    
        prop = getattr(cls, n)
        if not isinstance(prop, _TunableProperty):
            continue
        
        if prop._ntsubtable:
            key = '%s/%s/%s' % (prefix, prop._ntsubtable, n)
        else:
            key = '%s/%s' % (prefix, n)
        
        ntattr = '_Tunable__%s' % n
        
        ntvalue = NetworkTables.getGlobalAutoUpdateValue(key,
                                                         prop._ntdefault,
                                                         prop._ntwritedefault)
        # double indirection
        setattr(component, ntattr, ntvalue)
        prop._ntattr = ntattr


# TODO
#def autosend(f=None):
#    '''
#        Decorator used to send variables to DS::
#        
#            
#        
#            class MyComponent:
#            
#                @autosend
#                def my_sensor(self):
#                    return self.limit_switch.get()
#                    
#                ...
#                    
#            class MyRobot:
#            
#                mine = MyComponent 
#                
#        This will cause the output of this function to be sent
#        to ``/components/mine/my_sensor``
#    '''
#    
#    
#    def _get(self):
#        return self._Magicbot__autosend.get(f)
#    
#    prop = _AutosendProperty(fget=_get)
#    prop.f = f
#    
#    return prop



#----------------------------------------------------------------------------
# Copyright (c) Kauai Labs 2015. All Rights Reserved.
#
# Created in support of Team 2465 (Kauaibots).  Go Purple Wave!
#
# Open Source Software - may be modified and shared by FRC teams. Any
# modifications to this code must be accompanied by the \License.txt file
# in the root directory of the project
#----------------------------------------------------------------------------

import hal
import wpilib

import logging
logger = logging.getLogger('navx.i2c')

MAX_WPILIB_I2C_READ_BYTES = 127

class RegisterIO_I2C:
    
    def __init__(self, port):
        
        simPort = None
        if hal.HALIsSimulation():
            from ._impl.sim_io import NavXI2CSim
            simPort = NavXI2CSim()
        
        self.port = wpilib.I2C(port, 0x32, simPort=simPort)
        
    def init(self):
        logger.info("NavX i2c initialized")
        return True
    
    def write(self, address, value):
        return self.port.write(address | 0x80, value)
    
    def read(self, first_address, count):
        buffer = []
        l = 0
        while l != count:
            read_len = min(MAX_WPILIB_I2C_READ_BYTES, count - l)
            if not self.port.write(first_address + l, read_len):
                data = self.port.readOnly(read_len)
                if not data:
                    break
                
                buffer.extend(data)
                l = len(buffer)
            else:
                break
        
        if l != count:
            raise IOError("Read error")
        
        return buffer
    
    def shutdown(self):
        logger.info("NavX i2c shutdown")
        self.port.free()
        return True



#----------------------------------------------------------------------------
# Copyright (c) Kauai Labs 2015. All Rights Reserved.
#
# Created in support of Team 2465 (Kauaibots).  Go Purple Wave!
#
# Open Source Software - may be modified and shared by FRC teams. Any
# modifications to this code must be accompanied by the \License.txt file
# in the root directory of the project
#----------------------------------------------------------------------------

class IMURegisters:
    
    #********************************************
    # Device Identification Registers            
    #********************************************

    NAVX_REG_WHOAMI               = 0x00 # NAVX_MODEL_XXX 
    NAVX_REG_HW_REV               = 0x01
    NAVX_REG_FW_VER_MAJOR         = 0x02
    NAVX_REG_FW_VER_MINOR         = 0x03
    
    # Model types
    NAVX_MODEL_NAVX_MXP           = 0x32
    
    @classmethod
    def model_type(cls, whoami):
        if whoami == cls.NAVX_MODEL_NAVX_MXP:
            return "NavX MXP"
        else:
            return "unknown"

    #********************************************
    # Status and Control Registers               
    #********************************************

    # Read-write 
    NAVX_REG_UPDATE_RATE_HZ      = 0x04 # Range:  4 - 50 [unsigned byte] 
    # Read-only 
    # Accelerometer Full-Scale Range:  in units of G [unsigned byte] 
    NAVX_REG_ACCEL_FSR_G         = 0x05
    # Gyro Full-Scale Range (Degrees/Sec):  Range:  250, 500, 1000 or 2000 [unsigned short] 
    NAVX_REG_GYRO_FSR_DPS_L      = 0x06 # Lower 8-bits of Gyro Full-Scale Range 
    NAVX_REG_GYRO_FSR_DPS_H      = 0x07 # Upper 8-bits of Gyro Full-Scale Range 
    NAVX_REG_OP_STATUS           = 0x08 # NAVX_OP_STATUS_XXX 
    NAVX_REG_CAL_STATUS          = 0x09 # NAVX_CAL_STATUS_XXX 
    NAVX_REG_SELFTEST_STATUS     = 0x0A # NAVX_SELFTEST_STATUS_XXX 
    NAVX_REG_CAPABILITY_FLAGS_L  = 0x0B
    NAVX_REG_CAPABILITY_FLAGS_H  = 0x0C

    #********************************************
    # Processed Data Registers                   
    #********************************************

    NAVX_REG_SENSOR_STATUS_L     = 0x10 # NAVX_SENSOR_STATUS_XXX 
    NAVX_REG_SENSOR_STATUS_H     = 0x11
    # Timestamp:  [unsigned long] 
    NAVX_REG_TIMESTAMP_L_L       = 0x12
    NAVX_REG_TIMESTAMP_L_H       = 0x13
    NAVX_REG_TIMESTAMP_H_L       = 0x14
    NAVX_REG_TIMESTAMP_H_H       = 0x15

    # Yaw, Pitch, Roll:  Range: -180.00 to 180.00 [signed hundredths] 
    # Compass Heading:   Range: 0.00 to 360.00 [unsigned hundredths] 
    # Altitude in Meters:  In units of meters [16:16] 

    NAVX_REG_YAW_L               = 0x16 # Lower 8 bits of Yaw     
    NAVX_REG_YAW_H               = 0x17 # Upper 8 bits of Yaw     
    NAVX_REG_ROLL_L              = 0x18 # Lower 8 bits of Roll    
    NAVX_REG_ROLL_H              = 0x19 # Upper 8 bits of Roll    
    NAVX_REG_PITCH_L             = 0x1A # Lower 8 bits of Pitch   
    NAVX_REG_PITCH_H             = 0x1B # Upper 8 bits of Pitch   
    NAVX_REG_HEADING_L           = 0x1C # Lower 8 bits of Heading 
    NAVX_REG_HEADING_H           = 0x1D # Upper 8 bits of Heading 
    NAVX_REG_FUSED_HEADING_L     = 0x1E # Upper 8 bits of Fused Heading 
    NAVX_REG_FUSED_HEADING_H     = 0x1F # Upper 8 bits of Fused Heading 
    NAVX_REG_ALTITUDE_I_L        = 0x20
    NAVX_REG_ALTITUDE_I_H        = 0x21
    NAVX_REG_ALTITUDE_D_L        = 0x22
    NAVX_REG_ALTITUDE_D_H        = 0x23

    # World-frame Linear Acceleration: In units of +/- G * 1000 [signed thousandths] 

    NAVX_REG_LINEAR_ACC_X_L    = 0x24 # Lower 8 bits of Linear Acceleration X 
    NAVX_REG_LINEAR_ACC_X_H    = 0x25 # Upper 8 bits of Linear Acceleration X 
    NAVX_REG_LINEAR_ACC_Y_L    = 0x26 # Lower 8 bits of Linear Acceleration Y 
    NAVX_REG_LINEAR_ACC_Y_H    = 0x27 # Upper 8 bits of Linear Acceleration Y 
    NAVX_REG_LINEAR_ACC_Z_L    = 0x28 # Lower 8 bits of Linear Acceleration Z 
    NAVX_REG_LINEAR_ACC_Z_H    = 0x29 # Upper 8 bits of Linear Acceleration Z 

    # Quaternion:  Range -1 to 1 [signed short ratio] 

    NAVX_REG_QUAT_W_L             = 0x2A # Lower 8 bits of Quaternion W 
    NAVX_REG_QUAT_W_H             = 0x2B # Upper 8 bits of Quaternion W 
    NAVX_REG_QUAT_X_L             = 0x2C # Lower 8 bits of Quaternion X 
    NAVX_REG_QUAT_X_H             = 0x2D # Upper 8 bits of Quaternion X 
    NAVX_REG_QUAT_Y_L             = 0x2E # Lower 8 bits of Quaternion Y 
    NAVX_REG_QUAT_Y_H             = 0x2F # Upper 8 bits of Quaternion Y 
    NAVX_REG_QUAT_Z_L             = 0x30 # Lower 8 bits of Quaternion Z 
    NAVX_REG_QUAT_Z_H             = 0x31 # Upper 8 bits of Quaternion Z 

    #********************************************
    # Raw Data Registers                         
    #********************************************

    # Sensor Die Temperature:  Range +/- 150, In units of Centigrade * 100 [signed hundredths float 

    NAVX_REG_MPU_TEMP_C_L        = 0x32 # Lower 8 bits of Temperature 
    NAVX_REG_MPU_TEMP_C_H        = 0x33 # Upper 8 bits of Temperature 

    # Raw, Calibrated Angular Rotation, in device units.  Value in DPS = units / GYRO_FSR_DPS [signed short] 

    NAVX_REG_GYRO_X_L            = 0x34
    NAVX_REG_GYRO_X_H            = 0x35
    NAVX_REG_GYRO_Y_L            = 0x36
    NAVX_REG_GYRO_Y_H            = 0x37
    NAVX_REG_GYRO_Z_L            = 0x38
    NAVX_REG_GYRO_Z_H            = 0x39

    # Raw, Calibrated, Acceleration Data, in device units.  Value in G = units / ACCEL_FSR_G [signed short] 

    NAVX_REG_ACC_X_L            = 0x3A
    NAVX_REG_ACC_X_H            = 0x3B
    NAVX_REG_ACC_Y_L            = 0x3C
    NAVX_REG_ACC_Y_H            = 0x3D
    NAVX_REG_ACC_Z_L            = 0x3E
    NAVX_REG_ACC_Z_H            = 0x3F

    # Raw, Calibrated, Un-tilt corrected Magnetometer Data, in device units.  1 unit = 0.15 uTesla [signed short] 

    NAVX_REG_MAG_X_L            = 0x40
    NAVX_REG_MAG_X_H            = 0x41
    NAVX_REG_MAG_Y_L            = 0x42
    NAVX_REG_MAG_Y_H            = 0x43
    NAVX_REG_MAG_Z_L            = 0x44
    NAVX_REG_MAG_Z_H            = 0x45

    # Calibrated Pressure in millibars Valid Range:  10.00 Max:  1200.00 [16:16 float]  

    NAVX_REG_PRESSURE_IL       = 0x46
    NAVX_REG_PRESSURE_IH       = 0x47
    NAVX_REG_PRESSURE_DL       = 0x48
    NAVX_REG_PRESSURE_DH       = 0x49

    # Pressure Sensor Die Temperature:  Range +/- 150.00C [signed hundredths] 

    NAVX_REG_PRESSURE_TEMP_L    = 0x4A
    NAVX_REG_PRESSURE_TEMP_H    = 0x4B

    #********************************************
    # Calibration Registers                      
    #********************************************

    # Yaw Offset: Range -180.00 to 180.00 [signed hundredths] 

    NAVX_REG_YAW_OFFSET_L        = 0x4C # Lower 8 bits of Yaw Offset 
    NAVX_REG_YAW_OFFSET_H        = 0x4D # Upper 8 bits of Yaw Offset 

    # Quaternion Offset:  Range: -1 to 1 [signed short ratio]  

    NAVX_REG_QUAT_OFFSET_W_L     = 0x4E # Lower 8 bits of Quaternion W 
    NAVX_REG_QUAT_OFFSET_W_H     = 0x4F # Upper 8 bits of Quaternion W 
    NAVX_REG_QUAT_OFFSET_X_L     = 0x50 # Lower 8 bits of Quaternion X 
    NAVX_REG_QUAT_OFFSET_X_H     = 0x51 # Upper 8 bits of Quaternion X 
    NAVX_REG_QUAT_OFFSET_Y_L     = 0x52 # Lower 8 bits of Quaternion Y 
    NAVX_REG_QUAT_OFFSET_Y_H     = 0x53 # Upper 8 bits of Quaternion Y 
    NAVX_REG_QUAT_OFFSET_Z_L     = 0x54 # Lower 8 bits of Quaternion Z 
    NAVX_REG_QUAT_OFFSET_Z_H     = 0x55 # Upper 8 bits of Quaternion Z 

    #********************************************
    # Integrated Data Registers                  
    #********************************************

    # Integration Control (Write-Only)           
    NAVX_REG_INTEGRATION_CTL   = 0x56
    NAVX_REG_PAD_UNUSED        = 0x57

    # Velocity:  Range -32768.9999 - 32767.9999 in units of Meters/Sec      

    NAVX_REG_VEL_X_I_L         = 0x58
    NAVX_REG_VEL_X_I_H         = 0x59
    NAVX_REG_VEL_X_D_L         = 0x5A
    NAVX_REG_VEL_X_D_H         = 0x5B
    NAVX_REG_VEL_Y_I_L         = 0x5C
    NAVX_REG_VEL_Y_I_H         = 0x5D
    NAVX_REG_VEL_Y_D_L         = 0x5E
    NAVX_REG_VEL_Y_D_H         = 0x5F
    NAVX_REG_VEL_Z_I_L         = 0x60
    NAVX_REG_VEL_Z_I_H         = 0x61
    NAVX_REG_VEL_Z_D_L         = 0x62
    NAVX_REG_VEL_Z_D_H         = 0x63

    # Displacement:  Range -32768.9999 - 32767.9999 in units of Meters      

    NAVX_REG_DISP_X_I_L        = 0x64
    NAVX_REG_DISP_X_I_H        = 0x65
    NAVX_REG_DISP_X_D_L        = 0x66
    NAVX_REG_DISP_X_D_H        = 0x67
    NAVX_REG_DISP_Y_I_L        = 0x68
    NAVX_REG_DISP_Y_I_H        = 0x69
    NAVX_REG_DISP_Y_D_L        = 0x6A
    NAVX_REG_DISP_Y_D_H        = 0x6B
    NAVX_REG_DISP_Z_I_L        = 0x6C
    NAVX_REG_DISP_Z_I_H        = 0x6D
    NAVX_REG_DISP_Z_D_L        = 0x6E
    NAVX_REG_DISP_Z_D_H        = 0x6F
    
    NAVX_REG_LAST = NAVX_REG_DISP_Z_D_H




import hal
import time
import wpilib


class PreciseDelay:
    '''
        Used to synchronize a timing loop. Will delay precisely so that
        the next invocation of your loop happens at the same period, as long
        as your code does not run longer than the length of the delay.
        
        Our experience has shown that 25ms is a good loop period.
    
        Usage::
        
            delay = PreciseDelay(time_to_delay)
            
            while something:
                # do things here
                delay.wait()
    '''
    
    def __init__(self, delay_period):
        '''
            :param delay_period: The amount of time (in seconds) to do a delay
            :type delay_period: float
        '''
        
        # The WPILib sleep/etc functions are slightly less stable as
        # they have more overhead, so only use them in simulation mode
        if hal.HALIsSimulation():
            
            self.delay = wpilib.Timer.delay
            self.get_now = wpilib.Timer.getFPGATimestamp
            
            # Test to see if we're in a unit test, and switch the wait function
            # to run more efficiently -- otherwise full tests are dog slow
            try:
                import pyfrc.config
                if pyfrc.config.mode == 'test':
                    self.wait = self._wait_unit_tests
            except:
                pass
            
            
        else:
            self.delay = time.sleep
            self.get_now = time.time  
        
        self.delay_period = float(delay_period)
        if self.delay_period < 0.001:
            raise ValueError("You probably don't want to delay less than 1ms!")
        
        self.next_delay = self.get_now() + self.delay_period
    
    def wait(self):
        '''Waits until the delay period has passed'''
        
        # optimization -- avoid local lookups
        delay = self.delay
        get_now = self.get_now
        next_delay = self.next_delay
        
        while True:
            # we must *always* yield here, so other things can run
            delay(0.0002)
            
            if next_delay < get_now():
                break
        
        self.next_delay += self.delay_period

    def _wait_unit_tests(self):
        
        # Not optimized -- just need the unit tests to run fast
        # - TODO: should we just always use this in simulated mode?
        
        wpilib.Timer.delay(self.delay_period)
        
        



#----------------------------------------------------------------------------
# Copyright (c) Kauai Labs 2015. All Rights Reserved.
#
# Created in support of Team 2465 (Kauaibots).  Go Purple Wave!
#
# Open Source Software - may be modified and shared by FRC teams. Any
# modifications to this code must be accompanied by the \License.txt file
# in the root directory of the project
#----------------------------------------------------------------------------

DEFAULT_SPI_BITRATE_HZ = 500000

import hal
from robotpy_ext.misc.crc7 import crc7
from wpilib import SPI, Timer

import logging
logger = logging.getLogger('navx.spi')

class RegisterIO_SPI:
    
    def __init__(self, port, bitrate=None):
        
        if bitrate is None:
            bitrate = DEFAULT_SPI_BITRATE_HZ
        
        simPort = None
        if hal.HALIsSimulation():
            from ._impl.sim_io import NavXSPISim
            simPort = NavXSPISim()
        
        self.port = SPI(port, simPort=simPort)
        self.bitrate = bitrate
        
    def init(self):
        
        logger.info("Initializing NavX SPI")
        
        self.port.setClockRate(self.bitrate)
        self.port.setMSBFirst()
        self.port.setSampleDataOnFalling()
        self.port.setClockActiveLow()
        self.port.setChipSelectActiveLow()
        
        logger.info("Initialized!")
        return True
        
    def write(self, address, value):
        data = [address | 0x80, value]
        data.append(crc7(data))
        if self.port.write(data) != len(data):
            return False
        
        return True
        
    def read(self, first_address, count):
        
        data = [first_address, count]
        data.append(crc7(data))
        retcount = self.port.write(data)
        if retcount != len(data):
            raise IOError("Write error (%s != %s)" % (retcount, len(data)))
        
        Timer.delay(0.001)
        
        data = self.port.read(True, count + 1)
        if len(data) != count + 1:
            raise IOError("Read error (%s != %s)" % (len(data), count+1))
        
        crc = data[-1]
        data = data[:-1]
        
        if crc7(data) != crc:
            raise IOError("CRC error")
        
        return data
    
    def shutdown(self):
        # TODO: should free?
        logger.info("NavX i2c shutdown")
        self.port.free()
        return True




import contextlib
import inspect
import logging

import wpilib

from robotpy_ext.misc import PreciseDelay
from robotpy_ext.autonomous import AutonomousModeSelector

from robotpy_ext.misc.orderedclass import OrderedClass

from networktables import NetworkTable

from .magic_tunable import setup_tunables, _TunableProperty

__all__ = ['MagicRobot']

class MagicRobot(wpilib.SampleRobot,
                 metaclass=OrderedClass):
    """
        .. warning:: This implementation is still being developed, and may
                     be changed during the course of the 2016 season.

        Robots that use the MagicBot framework should use this as their
        base robot class. If you use this as your base, you must
        implement the following methods:

        - :meth:`createObjects`
        - :meth:`teleopPeriodic`

        MagicRobot uses the :class:`.AutonomousModeSelector` to allow you
        to define multiple autonomous modes and to select one of them via
        the SmartDashboard/SFX.
        
        MagicRobot will set the following NetworkTables variables
        automatically:
        
        - ``/robot/mode``: one of 'disabled', 'auto', 'teleop', or 'test'
        - ``/robot/is_simulation``: True/False
        - ``/robot/is_ds_attached``: True/False
        
    """

    #: Amount of time each loop takes (default is 20ms)
    control_loop_wait_time = 0.020

    #: Error report interval: when an FMS is attached, how often should
    #: uncaught exceptions be reported?
    error_report_interval = 0.5

    def robotInit(self):
        """
            .. warning:: Internal API, don't override; use :meth:`createObjects` instead
        """

        self._exclude_from_injection = [
            'logger', 'members'
        ]

        self.__last_error_report = -10

        self._components = []

        # Create the user's objects and stuff here
        self.createObjects()

        # Load autonomous modes
        self._automodes = AutonomousModeSelector('autonomous')

        # Next, create the robot components and wire them together
        self._create_components()
        
        self.__nt = NetworkTable.getTable('/robot')
        self.__nt.putBoolean('is_simulation', self.isSimulation())
        self.__nt.putBoolean('is_ds_attached', self.ds.isDSAttached())

    def createObjects(self):
        """
            You should override this and initialize all of your wpilib
            objects here (and not in your components, for example). This
            serves two purposes:

            - It puts all of your motor/sensor initialization in the same
              place, so that if you need to change a port/pin number it
              makes it really easy to find it. Additionally, if you want
              to create a simplified robot program to test a specific
              thing, it makes it really easy to copy/paste it elsewhere

            - It allows you to use the magic injection mechanism to share
              variables between components

            .. note:: Do not access your magic components in this function,
                      as their instances have not been created yet. Do not
                      create them either.
        """
        raise NotImplementedError

    def teleopInit(self):
        """
            Initialization code for teleop control code may go here.

            Users may override this method for initialization code which will be
            called each time the robot enters teleop mode.

            .. note:: The ``on_enable`` functions of all components are called
                      before this function is called.
        """
        pass

    def teleopPeriodic(self):
        """
            Periodic code for teleop mode should go here.

            Users should override this method for code which will be called
            periodically at a regular rate while the robot is in teleop mode.

            This code executes before the ``execute`` functions of all
            components are called.
        """
        func = self.teleopPeriodic.__func__
        if not hasattr(func, "firstRun"):
            self.logger.warn("Default MagicRobot.teleopPeriodic() method... Overload me!")
            func.firstRun = False

    def disabledInit(self):
        """
            Initialization code for disabled mode may go here.

            Users may override this method for initialization code which will be
            called each time the robot enters disabled mode.

            .. note:: The ``on_disable`` functions of all components are called
                      before this function is called.
        """
        pass

    def disabledPeriodic(self):
        """
            Periodic code for disabled mode should go here.

            Users should override this method for code which will be called
            periodically at a regular rate while the robot is in disabled mode.

            This code executes before the ``execute`` functions of all
            components are called.
        """
        func = self.disabledPeriodic.__func__
        if not hasattr(func, "firstRun"):
            self.logger.warn("Default MagicRobot.disabledPeriodic() method... Overload me!")
            func.firstRun = False

    def onException(self, forceReport=False):
        '''
            This function must *only* be called when an unexpected exception
            has occurred that would otherwise crash the robot code. Use this
            inside your :meth:`operatorActions` function.

            If the FMS is attached (eg, during a real competition match),
            this function will return without raising an error. However,
            it will try to report one-off errors to the Driver Station so
            that it will be recorded in the Driver Station Log Viewer.
            Repeated errors may not get logged.

            Example usage::

                def teleopPeriodic(self):
                    try:
                        if self.joystick.getTrigger():
                            self.shooter.shoot()
                    except:
                        self.onException()

                    try:
                        if self.joystick.getRawButton(2):
                            self.ball_intake.run()
                    except:
                        self.onException()

                    # and so on...

            :param forceReport: Always report the exception to the DS. Don't
                                set this to True
        '''
        # If the FMS is not attached, crash the robot program
        if not self.ds.isFMSAttached():
            raise

        # Otherwise, if the FMS is attached then try to report the error via
        # the driver station console. Maybe.
        now = wpilib.Timer.getFPGATimestamp()

        try:
            if forceReport or (now - self.__last_error_report) > self.error_report_interval:
                wpilib.DriverStation.reportError("Unexpected exception", True)
        except:
            pass    # ok, can't do anything here

        self.__last_error_report = now

    @contextlib.contextmanager
    def consumeExceptions(self, forceReport=False):
        """
            This returns a context manager which will consume any uncaught
            exceptions that might otherwise crash the robot.

            Example usage::

                def teleopPeriodic(self):
                    with self.consumeExceptions():
                        if self.joystick.getTrigger():
                            self.shooter.shoot()

                    with self.consumeExceptions():
                        if self.joystick.getRawButton(2):
                            self.ball_intake.run()

                    # and so on...

            :param forceReport: Always report the exception to the DS. Don't
                                set this to True

            .. seealso:: :meth:`onException` for more details
        """
        try:
            yield
        except:
            self.onException(forceReport=forceReport)

    #
    # Internal API
    #

    def autonomous(self):
        """
            MagicRobot will do The Right Thing and automatically load all
            autonomous mode routines defined in the autonomous folder.

            .. warning:: Internal API, don't override
        """
        
        self.__nt.putString('mode', 'auto')
        self.__nt.putBoolean('is_ds_attached', self.ds.isDSAttached())

        self._on_mode_enable_components()

        self._automodes.run(self.control_loop_wait_time,
                            self._execute_components,
                            self.onException)

        self._on_mode_disable_components()


    def disabled(self):
        """
            This function is called in disabled mode. You should not
            override this function; rather, you should override the
            :meth:`disabledPeriodic` function instead.

            .. warning:: Internal API, don't override
        """
        
        self.__nt.putString('mode', 'disabled')
        ds_attached = None

        delay = PreciseDelay(self.control_loop_wait_time)

        self._on_mode_disable_components()
        try:
            self.disabledInit()
        except:
            self.onException(forceReport=True)

        while self.isDisabled():
            
            if ds_attached != self.ds.isDSAttached():
                ds_attached = not ds_attached
                self.__nt.putBoolean('is_ds_attached', ds_attached)
            
            try:
                self.disabledPeriodic()
            except:
                self.onException()

            delay.wait()

    def operatorControl(self):
        """
            This function is called in teleoperated mode. You should not
            override this function; rather, you should override the
            :meth:`teleopPeriodics` function instead.

            .. warning:: Internal API, don't override
        """
        
        self.__nt.putString('mode', 'teleop')
        # don't need to update this during teleop -- presumably will switch
        # modes when ds is no longer attached
        self.__nt.putBoolean('is_ds_attached', self.ds.isDSAttached())

        delay = PreciseDelay(self.control_loop_wait_time)

        # initialize things
        self._on_mode_enable_components()

        try:
            self.teleopInit()
        except:
            self.onException(forceReport=True)

        while self.isOperatorControl() and self.isEnabled():
            
            #self._update_autosend()
            
            try:
                self.teleopPeriodic()
            except:
                self.onException()

            self._execute_components()
            delay.wait()

        self._on_mode_disable_components()

    def test(self):
        '''Called when the robot is in test mode'''
        
        self.__nt.putString('mode', 'test')
        self.__nt.putBoolean('is_ds_attached', self.ds.isDSAttached())
        
        while self.isTest() and self.isEnabled():
            wpilib.LiveWindow.run()
            wpilib.Timer.delay(.01)

    def _on_mode_enable_components(self):
        # initialize things
        for component in self._components:
            if hasattr(component, 'on_enable'):
                try:
                    component.on_enable()
                except:
                    self.onException(forceReport=True)

    def _on_mode_disable_components(self):
        # deinitialize things
        for component in self._components:
            if hasattr(component, 'on_disable'):
                try:
                    component.on_disable()
                except:
                    self.onException(forceReport=True)

    def _create_components(self):

        #
        # TODO: Will need to inject into any autonomous mode component
        #       too, as they're a bit different
        #

        # TODO: Will need to order state machine components before
        #       other components just in case

        components = []
        
        self.logger.info("Creating magic components")

        # Identify all of the types, and create them
        cls = self.__class__

        for m in self.members:
            if m.startswith('_') or isinstance(getattr(cls, m, None), _TunableProperty):
                continue
            
            ctyp = getattr(self, m)
            if not isinstance(ctyp, type):
                continue

            # Create instance, set it on self
            component = ctyp()
            setattr(self, m, component)

            # Ensure that mandatory methods are there
            if not hasattr(component, 'execute') or \
               not callable(component.execute):
                raise ValueError("Component %s (%s) must have a method named 'execute'" % (m, component))

            # Automatically inject a logger object
            component.logger = logging.getLogger(m)
            component._Magicbot__autosend = {}

            # Store for later
            components.append((m, component))
            
            self.logger.info("-> %s (class: %s)", m, ctyp.__name__)

        # Collect attributes of this robot that are injectable
        self._injectables = {}

        for n in dir(self):
            if n.startswith('_') or n in self._exclude_from_injection or \
               isinstance(getattr(cls, n, None), _TunableProperty):
                continue

            o = getattr(self, n)

            # Don't inject methods
            # TODO: This could actually be a cool capability..
            if inspect.ismethod(o):
                continue

            self._injectables[n] = o

        # For each new component, perform magic injection
        for cname, component in components:
            self._components.append(component)
            setup_tunables(component, cname, 'components')
            self._setup_vars(cname, component)

        # Do it for autonomous modes too
        for mode in self._automodes.modes.values():
            mode.logger = logging.getLogger(mode.MODE_NAME)
            setup_tunables(mode, mode.MODE_NAME, 'autonomous')
            self._setup_vars(mode.MODE_NAME, mode)

        # And for self too
        setup_tunables(self, 'robot', None)
            
        # Call setup functions for components
        for cname, component in components:
            if hasattr(component, 'setup'):
                component.setup()
    
    def _setup_vars(self, cname, component):
        
        self.logger.debug("Injecting magic variables into %s", cname)
        
        for n in dir(component):
            if n.startswith('_') or isinstance(getattr(type(component), n, True), property):
                continue
            
            inject_type = getattr(component, n)
            
            if not isinstance(inject_type, type):
                continue

            injectable = self._injectables.get(n)
            if injectable is None:
                if cname is not None:
                    injectable = self._injectables.get('%s_%s' % (cname, n))

            if injectable is None:
                raise ValueError("Component %s has variable %s (type %s), which is not present in %s" %
                                 (cname, n, inject_type, self))

            if not isinstance(injectable, inject_type):
                raise ValueError("Component %s variable %s in %s are not the same types! (Got %s, expected %s)" %
                                 (cname, n, self, type(injectable), inject_type))

            setattr(component, n, injectable)
            self.logger.debug("-> %s as %s.%s", injectable, cname, n)
        
        # XXX
        #if is_autosend:
            # where to store the nt key?
        #    component._Magicbot__autosend[prop.f] = None
    
    #def _update_autosend(self):
    #    # seems like this should just be a giant list instead
    #    for component in self._components:
    #        d = component._Magicbot__autosend
    #        for f in d.keys():
    #            d[f] = f(component)     

    def _execute_components(self):
        for component in self._components:
            try:
                component.execute()
            except:
                self.onException()



#----------------------------------------------------------------------------
# Copyright (c) Kauai Labs 2015. All Rights Reserved.
#
# Created in support of Team 2465 (Kauaibots).  Go Purple Wave!
#
# Open Source Software - may be modified and shared by FRC teams. Any
# modifications to this code must be accompanied by the \License.txt file
# in the root directory of the project
#----------------------------------------------------------------------------

class ContinuousAngleTracker:
    
    def __init__(self):
        self.last_angle = 0.0
        self.zero_crossing_count = 0
        self.last_rate = 0
        self.first_sample = True
    
    def nextAngle(self, newAngle):
        
        # If the first received sample is negative, 
        # ensure that the zero crossing count is
        # decremented.
        
        if self.first_sample:
            self.first_sample = False
            if newAngle < 0.0:
                self.zero_crossing_count -= 1
        
        # Calculate delta angle, adjusting appropriately
        # if the current sample crossed the -180/180
        # point.
        
        
        bottom_crossing = False
        delta_angle = newAngle - self.last_angle
        
        # Adjust for wraparound at -180/+180 point
        if delta_angle >= 180.0:
            delta_angle = 360.0 - delta_angle
            bottom_crossing = True
        elif delta_angle <= -180.0:
            delta_angle = 360.0 + delta_angle;
            bottom_crossing = True
        
        self.last_rate = delta_angle

        # If a zero crossing occurred, increment/decrement
        # the zero crossing count appropriately.
        if not bottom_crossing:
            if delta_angle < 0.0:
                if (newAngle < 0.0) and (self.last_angle >= 0.0):
                    self.zero_crossing_count -= 1
            elif delta_angle >= 0.0:
                if (newAngle >= 0.0) and (self.last_angle < 0.0):
                    self.zero_crossing_count += 1
        
        self.last_angle = newAngle
    
    def getAngle(self):
        accumulated_angle = self.zero_crossing_count * 360.0
        curr_angle = self.last_angle
        if curr_angle < 0.0:
            curr_angle += 360.0
        
        accumulated_angle += curr_angle
        return accumulated_angle
    
    def getRate(self):
        return self.last_rate



import wpilib
import math

class SharpIR2Y0A02:
    '''
        Sharp GP2Y0A02YK0F is an analog IR sensor capable of measuring
        distances from 20cm to 150cm. Output distance is measured in
        centimeters.
        
        Distance is calculated using the following equation derived from
        the graph provided in the datasheet::
        
            62.28*x ^ -1.092
            
        .. warning:: FRC Teams: the case on these sensors is conductive and
                     grounded, and should not be mounted on a metallic
                     surface!
    '''
    
    def __init__(self, port):
        ''':param port: Analog port number'''
        self.distance = wpilib.AnalogInput(port)
        
    def getDistance(self):
        '''
            :returns: distance in centimeters. The output is constrained to
                      be between 22.5 and 145
        '''
        
        # Don't allow zero/negative values
        v = max(self.distance.getVoltage(), 0.00001)
        d = 62.28*math.pow(v, -1.092)
        
        # Constrain output 
        return max(min(d, 145.0), 22.5)


class SharpIRGP2Y0A41SK0F:
    '''
        Sharp GP2Y0A41SK0F is an analog IR sensor capable of measuring
        distances from 4cm to 40cm. Output distance is measured in
        centimeters.
    
        Distance is calculated using the following equation derived from
        the graph provided in the datasheet::
        
            12.84*x ^ -0.9824
            
        .. warning:: FRC Teams: the case on these sensors is conductive and
                     grounded, and should not be mounted on a metallic
                     surface!
    '''
    
    def __init__(self, port):
        ''':param port: Analog port number'''
        self.distance = wpilib.AnalogInput(port)

    def getDistance(self):
        '''
            :returns: distance in centimeters. The output is constrained to
                      be between 4.5 and 35
        '''
        
        # Don't allow zero/negative values
        v = max(self.distance.getVoltage(), 0.00001)
        d = 12.84*math.pow(v, -0.9824)
        
        # Constrain output 
        return max(min(d, 35.0), 4.5)



#----------------------------------------------------------------------------
# Copyright (c) Kauai Labs 2015. All Rights Reserved.
#
# Created in support of Team 2465 (Kauaibots).  Go Purple Wave!
#
# Open Source Software - may be modified and shared by FRC teams. Any
# modifications to this code must be accompanied by the \License.txt file
# in the root directory of the project
#----------------------------------------------------------------------------

class IMUProtocol:
    PACKET_START_CHAR =        '!'
    PROTOCOL_FLOAT_LENGTH =            7
    CHECKSUM_LENGTH =                  2
    TERMINATOR_LENGTH =                2

    # Yaw/Pitch/Roll (YPR) Update Packet - e.g., !y[yaw][pitch][roll][compass_heading]
    MSGID_YPR_UPDATE =                 'y'
    YPR_UPDATE_YAW_VALUE_INDEX =               2
    YPR_UPDATE_PITCH_VALUE_INDEX =             9
    YPR_UPDATE_ROLL_VALUE_INDEX =              16
    YPR_UPDATE_COMPASS_VALUE_INDEX =           23
    YPR_UPDATE_CHECKSUM_INDEX =                30
    YPR_UPDATE_TERMINATOR_INDEX =              32
    YPR_UPDATE_MESSAGE_LENGTH =                34

    # Quaternion Data Update Packet - e.g., !r[q1][q2][q3][q4][accelx][accely][accelz][magx][magy][magz]
    MSGID_QUATERNION_UPDATE =          'q'
    QUATERNION_UPDATE_MESSAGE_LENGTH =         53
    QUATERNION_UPDATE_QUAT1_VALUE_INDEX =      2
    QUATERNION_UPDATE_QUAT2_VALUE_INDEX =      6
    QUATERNION_UPDATE_QUAT3_VALUE_INDEX =      10
    QUATERNION_UPDATE_QUAT4_VALUE_INDEX =      14
    QUATERNION_UPDATE_ACCEL_X_VALUE_INDEX =    18
    QUATERNION_UPDATE_ACCEL_Y_VALUE_INDEX =    22
    QUATERNION_UPDATE_ACCEL_Z_VALUE_INDEX =    26
    QUATERNION_UPDATE_MAG_X_VALUE_INDEX =      30
    QUATERNION_UPDATE_MAG_Y_VALUE_INDEX =      34
    QUATERNION_UPDATE_MAG_Z_VALUE_INDEX =      38
    QUATERNION_UPDATE_TEMP_VALUE_INDEX =       42
    QUATERNION_UPDATE_CHECKSUM_INDEX =         49
    QUATERNION_UPDATE_TERMINATOR_INDEX =       51

    # Gyro/Raw Data Update packet - e.g., !g[gx][gy][gz][accelx][accely][accelz][magx][magy][magz][temp_c][cr][lf]

    MSGID_GYRO_UPDATE =                'g'
    GYRO_UPDATE_GYRO_X_VALUE_INDEX =           2
    GYRO_UPDATE_GYRO_Y_VALUE_INDEX =           6
    GYRO_UPDATE_GYRO_Z_VALUE_INDEX =           10
    GYRO_UPDATE_ACCEL_X_VALUE_INDEX =          14
    GYRO_UPDATE_ACCEL_Y_VALUE_INDEX =          18
    GYRO_UPDATE_ACCEL_Z_VALUE_INDEX =          22
    GYRO_UPDATE_MAG_X_VALUE_INDEX =            26
    GYRO_UPDATE_MAG_Y_VALUE_INDEX =            30
    GYRO_UPDATE_MAG_Z_VALUE_INDEX =            34
    GYRO_UPDATE_TEMP_VALUE_INDEX =             38
    GYRO_UPDATE_CHECKSUM_INDEX =               42
    GYRO_UPDATE_TERMINATOR_INDEX =             44
    GYRO_UPDATE_MESSAGE_LENGTH =               46
    
    # EnableStream Command Packet - e.g., !S[stream type][checksum][cr][lf]
    MSGID_STREAM_CMD =                 'S'
    STREAM_CMD_STREAM_TYPE_YPR =        MSGID_YPR_UPDATE
    STREAM_CMD_STREAM_TYPE_QUATERNION = MSGID_QUATERNION_UPDATE
    STREAM_CMD_STREAM_TYPE_GYRO =       MSGID_GYRO_UPDATE
    STREAM_CMD_STREAM_TYPE_INDEX =             2
    STREAM_CMD_UPDATE_RATE_HZ_INDEX =          3
    STREAM_CMD_CHECKSUM_INDEX =                5
    STREAM_CMD_TERMINATOR_INDEX =              7
    STREAM_CMD_MESSAGE_LENGTH =                9

    # EnableStream Response Packet - e.g., !s[stream type][gyro full scale range][accel full scale range][update rate hz][yaw_offset_degrees][flags][checksum][cr][lf]
    MSG_ID_STREAM_RESPONSE =           's'
    STREAM_RESPONSE_MESSAGE_LENGTH =           46
    STREAM_RESPONSE_STREAM_TYPE_INDEX =        2
    STREAM_RESPONSE_GYRO_FULL_SCALE_DPS_RANGE = 3
    STREAM_RESPONSE_ACCEL_FULL_SCALE_G_RANGE = 7
    STREAM_RESPONSE_UPDATE_RATE_HZ =           11
    STREAM_RESPONSE_YAW_OFFSET_DEGREES =       15
    STREAM_RESPONSE_QUAT1_OFFSET =             22
    STREAM_RESPONSE_QUAT2_OFFSET =             26
    STREAM_RESPONSE_QUAT3_OFFSET =             30
    STREAM_RESPONSE_QUAT4_OFFSET =             34
    STREAM_RESPONSE_FLAGS        =             38
    STREAM_RESPONSE_CHECKSUM_INDEX =           42
    STREAM_RESPONSE_TERMINATOR_INDEX =         44
    
    STREAM_MSG_TERMINATION_CHAR = '\n'
    
    NAV6_FLAG_MASK_CALIBRATION_STATE =    0x03
    
    NAV6_CALIBRATION_STATE_WAIT =         0x00
    NAV6_CALIBRATION_STATE_ACCUMULATE =   0x01
    NAV6_CALIBRATION_STATE_COMPLETE =     0x02
    
    IMU_PROTOCOL_MAX_MESSAGE_LENGTH = QUATERNION_UPDATE_MESSAGE_LENGTH
    
    class GyroUpdate:
        raw_gyro_x = 0
        raw_gyro_y = 0
        raw_gyro_z = 0
        raw_accel_x = 0
        raw_accel_y = 0
        raw_accel_z = 0
        cal_mag_x = 0
        cal_mag_y = 0
        cal_mag_z = 0
        mpu_temp_c = 0.0
        

    



from glob import glob
import importlib
import inspect
import os

import logging
logger = logging.getLogger('autonomous')

from ..misc.precise_delay import PreciseDelay

import wpilib


class AutonomousModeSelector:
    '''
        This object loads all modules in a specified python package, and tries
        to automatically discover autonomous modes from them. Each module is 
        added to a ``SendableChooser`` object, which allows the user to select
        one of them via the SmartDashboard/SFX.
        
        Autonomous mode objects must implement the following functions:
        
        - ``on_enable`` - Called when autonomous mode is initially enabled
        - ``on_disable`` - Called when autonomous mode is no longer active
        - ``on_iteration`` - Called for each iteration of the autonomous control loop
        
        Your autonomous object may have the following attributes:
        
        - ``MODE_NAME`` - The name of the autonomous mode to display to users
        - ``DISABLED`` - If True, don't allow this mode to be selected
        - ``DEFAULT`` - If True, this is the default autonomous mode selected
        
        AutonomousModeSelector can be easily used from either ``IterativeRobot``
        or from ``SampleRobot``. For ``IterativeRobot``, the simplest usage would
        look like so::
        
            class MyRobot(wpilib.IterativeRobot):
            
                def robotInit(self):
                    self.automodes = AutonomousModeSelector('autonomous')
            
                def autonomousPeriodic(self):
                    self.automodes.run()
                    
        
        For ``SampleRobot``, for the simplest usage you would do this::
        
            class MyRobot(wpilib.SampleRobot):
            
                def robotInit(self):
                    self.automodes = AutonomousModeSelector('autonomous')
            
                def autonomous(self):
                    self.automodes.run()
       
        If you use AutonomousModeSelector, you may also be interested in
        the autonomous state machine helper (:class:`.StatefulAutonomous`).
        
        Check out the samples in our github repository that show some basic
        usage of ``AutonomousModeSelector``.
        
        .. note:: If you use AutonomousModeSelector, then you should add
                  ``robotpy_ext.autonomous.selector_tests`` to your pyfrc
                  unit tests like so::
                  
                      from robotpy_ext.autonomous.selector_tests import *
    '''
    
    def __init__(self, autonomous_pkgname, *args, **kwargs):
        '''
            :param autonomous_pkgname: Module to load autonomous modes from
            :param args: Args to pass to created autonomous modes
            :param kwargs: Keyword args to pass to created autonomous modes
        '''
        
        self.ds = wpilib.DriverStation.getInstance()
        self.modes = {}
        self.active_mode = None
        
        logger.info("Begin initializing autonomous mode switcher")
        
        # load all modules in specified module
        modules = []
        
        try:
            autonomous_pkg = importlib.import_module(autonomous_pkgname)
        except ImportError as e:
            if e.name not in [autonomous_pkgname, autonomous_pkgname.split('.')[0]]:
                raise
            
            # Don't kill the robot because they didn't create an autonomous package
            logger.warn("Cannot load the '%s' package", autonomous_pkgname)
        else:
            if hasattr(autonomous_pkg, '__file__'):
                modules_path = os.path.dirname(os.path.abspath(autonomous_pkg.__file__))
                modules = glob(os.path.join(modules_path, '*.py' ))
        
        for module_filename in modules:
            
            module = None
            module_name = os.path.basename(module_filename[:-3])
            
            if module_name in  ['__init__']:
                continue
        
            try:
                module = importlib.import_module('.' + module_name, autonomous_pkgname)
                #module = imp.load_source('.' + module_name, module_filename)
            except:
                if not self.ds.isFMSAttached():
                    raise
            
            #
            # Find autonomous mode classes in the modules that are present
            # -> note that we actually create the instance of the objects here,
            #    so that way we find out about any errors *before* we get out 
            #    on the field.. 
            
            for name, obj in inspect.getmembers(module, inspect.isclass):

                if hasattr(obj, 'MODE_NAME') :
                    
                    # don't allow the driver to select this mode 
                    if hasattr(obj, 'DISABLED') and obj.DISABLED:
                        logger.warn("autonomous mode %s is marked as disabled", obj.MODE_NAME)
                        continue
                    
                    try:
                        instance = obj(*args, **kwargs)
                    except:
                        
                        if not self.ds.isFMSAttached():
                            raise
                        else:
                            continue
                    
                    if instance.MODE_NAME in self.modes:
                        if not self.ds.isFMSAttached():
                            raise RuntimeError( "Duplicate name %s in %s" % (instance.MODE_NAME, module_filename) )
                        
                        logger.error("Duplicate name %s specified by object type %s in module %s", instance.MODE_NAME, name, module_filename)
                        self.modes[name + '_' + module_filename] = instance
                    else:
                        self.modes[instance.MODE_NAME] = instance
        
        # now that we have a bunch of valid autonomous mode objects, let 
        # the user select one using the SmartDashboard.
        
        # SmartDashboard interface
        self.chooser = wpilib.SendableChooser()
        
        default_modes = []
        
        logger.info("Loaded autonomous modes:")
        for k,v in sorted(self.modes.items()):
            
            if hasattr(v, 'DEFAULT') and v.DEFAULT == True:
                logger.info(" -> %s [Default]", k)
                self.chooser.addDefault(k, v)
                default_modes.append(k)
            else:
                logger.info( " -> %s", k )
                self.chooser.addObject(k, v)
        
        if len(self.modes) == 0:
            logger.warn("-- no autonomous modes were loaded!")
                
        self.chooser.addObject('None', None)
        
        if len(default_modes) == 0:
            self.chooser.addDefault('None', None)
        elif len(default_modes) != 1:
            if not self.ds.isFMSAttached():
                raise RuntimeError("More than one autonomous mode was specified as default! (modes: %s)" % (', '.join(default_modes)))
            
                
        # must PutData after setting up objects
        wpilib.SmartDashboard.putData('Autonomous Mode', self.chooser)
        
        logger.info("Autonomous switcher initialized")
    
            
    def run(self, control_loop_wait_time=0.020, iter_fn=None, on_exception=None):    
        '''
            This function does everything required to implement autonomous
            mode behavior. You should call this from your autonomous mode
            function -- ``autonomousPeriodic`` in :class:`.IterativeRobot`,
            or ``autonomous`` in :class:`.SampleRobot`.
            
            This function will NOT exit until autonomous mode has ended. If
            you need to execute code in all autonomous modes, pass a function
            as the ``iter_fn`` parameter, and it will be called once per
            autonomous mode iteration.
                          
            :param control_loop_wait_time: Amount of time between iterations
            :param iter_fn: Called at the end of every iteration while
                            autonomous mode is executing
            :param on_exception: Called when an uncaught exception is raised,
                                 must take a single keyword arg "forceReport"
        '''
        
        logger.info("Begin autonomous")
        
        if iter_fn is None:
            iter_fn = lambda: None
            
        if on_exception is None:
            on_exception = self._on_exception
        
        # keep track of how much time has passed in autonomous mode
        timer = wpilib.Timer()
        timer.start()
        
        try:
            self._on_autonomous_enable()
        except:
            on_exception(forceReport=True)
        
        #
        # Autonomous control loop
        #
        
        delay = PreciseDelay(control_loop_wait_time)
        
        while self.ds.isAutonomous() and self.ds.isEnabled():
 
            try:            
                self._on_iteration(timer.get())
            except:
                on_exception()
            
            iter_fn()
             
            delay.wait()
            
        #
        # Done with autonomous, finish up
        #
            
        try:
            self._on_autonomous_disable()
        except:
            on_exception(forceReport=True)
            
        logger.info("Autonomous mode ended")
        
    #
    #   Internal methods used to implement autonomous mode switching, and
    #   are called automatically
    #
    
    def _on_autonomous_enable(self):
        '''Selects the active autonomous mode and enables it'''
        self.active_mode = self.chooser.getSelected()
        if self.active_mode is not None:
            logger.info("Enabling '%s'" % self.active_mode.MODE_NAME)
            self.active_mode.on_enable()
        else:
            logger.warn("No autonomous modes were selected, not enabling autonomous mode")
 
    def _on_autonomous_disable(self):
        '''Disable the active autonomous mode'''
        if self.active_mode is not None:
            logger.info("Disabling '%s'" % self.active_mode.MODE_NAME)
            self.active_mode.on_disable()
            
        self.active_mode = None
        
    def _on_iteration(self, time_elapsed):
        '''Run the code for the current autonomous mode'''
        if self.active_mode is not None:
            self.active_mode.on_iteration(time_elapsed)
            
    def _on_exception(self, forceReport=False):
        if not self.ds.isFMSAttached():
            raise




#----------------------------------------------------------------------------
# Copyright (c) Kauai Labs 2015. All Rights Reserved.
#
# Created in support of Team 2465 (Kauaibots).  Go Purple Wave!
#
# Open Source Software - may be modified and shared by FRC teams. Any
# modifications to this code must be accompanied by the \License.txt file
# in the root directory of the project
#----------------------------------------------------------------------------

from collections import deque

class OffsetTracker:
    
    def __init__(self, history_length):
        self.value_history = deque([0]*history_length, maxlen=history_length)
        self.value_offset = 0
    
    def updateHistory(self, value):
        self.value_history.append(value)

    def getAverageFromHistory(self):
        return sum(self.value_history) / float(len(self.value_history))
    
    def setOffset(self):
        self.value_offset = self.getAverageFromHistory()
    
    def getOffset(self):
        return self.value_offset
    
    def applyOffset(self, value):
        offseted_value = (value - self.value_offset)
        if offseted_value < -180:
            offseted_value += 360
        
        if offseted_value > 180:
            offseted_value -= 360
        
        return offseted_value



#----------------------------------------------------------------------------
# Copyright (c) Kauai Labs 2015. All Rights Reserved.
#
# Created in support of Team 2465 (Kauaibots).  Go Purple Wave!
#
# Open Source Software - may be modified and shared by FRC teams. Any
# modifications to this code must be accompanied by the \License.txt file
# in the root directory of the project
#----------------------------------------------------------------------------


class InertialDataIntegrator:
    
    def __init__(self):
        self.resetDisplacement()
    
    def updateDisplacement(self, accel_x_g, accel_y_g,
                           update_rate_hz, is_moving):
        if is_moving:
            accel_g = (accel_x_g, accel_y_g)
            sample_time = 1.0 / update_rate_hz
            
            for i in range(2):
                m_s2 = accel_g[i] * 9.80665
                self.displacement[i] += self.last_velocity[i] + (0.5 * m_s2 * sample_time * sample_time)
                self.last_velocity[i] = self.last_velocity[i] + (m_s2 * sample_time)
        else:
            self.last_velocity[0] = 0
            self.last_velocity[1] = 0
            
    def resetDisplacement(self):
        self.displacement = [0, 0]
        self.last_velocity = [0, 0]
    
    def getVelocityX(self):
        return self.last_velocity[0]
    
    def getVelocityY(self):
        return self.last_velocity[1]
    
    def getVelocityZ(self):
        return 0
    
    def getDisplacementX(self):
        return self.displacement[0]
    
    def getDisplacementY(self):
        return self.displacement[1]
    
    def getDisplacementZ(self):
        return 0



import warnings

import hal
import wpilib

class XboxController:
    '''
        Allows usage of an Xbox controller, with sensible names for xbox
        specific buttons and axes.
    
        Mapping based on http://www.team358.org/files/programming/ControlSystem2015-2019/images/XBoxControlMapping.jpg

        :deprecated: Use :class:`wpilib.XboxController <wpilib.xboxcontroller.XboxController>` instead
    '''
    
    def __init__(self, port):
        '''
        :param port: The port on the driver station that the controller is
            plugged into.
        :type  port: int
        '''
    
        warnings.warn('robotpy_ext.control.xbox_controller is deprecated, use wpilib.XboxController instead', category=DeprecationWarning, stacklevel=2)

        self.ds = wpilib.DriverStation.getInstance()
        self.port = port
        
        hal.report(hal.HALUsageReporting.kResourceType_Joystick, port)
    
    def getLeftX(self):
        """Get the left stick X axis
        
        :returns: -1 to 1
        :rtype: float
        """
        return self.ds.getStickAxis(self.port, 0)
    
    def getLeftY(self):
        """Get the left stick Y axis
        
        :returns: -1 to 1
        :rtype: float
        """
        return self.ds.getStickAxis(self.port, 1)
    
    getX = getLeftX
    getY = getLeftY
    
    def getLeftPressed(self):
        """Determines if the left stick is pressed
        
        :returns: True if pressed, False otherwise
        :rtype: bool
        """
        return self.ds.getStickButton(self.port, 8)
    
    def getPOV(self):
        """Get the state of a POV on the joystick.

        :param pov: which POV (default is 0)
        :type  pov: int
        :returns: The angle of the POV in degrees, or -1 if the POV is not
                  pressed.
        :rtype: float
        """
        return self.ds.getStickPOV(self.port, 0)
    
    def getRightX(self):
        """Get the right stick X axis
        
        :returns: -1 to 1
        :rtype: float
        """
        return self.ds.getStickAxis(self.port, 4)
    
    def getRightY(self):
        """Get the right stick Y axis
        
        :returns: -1 to 1
        :rtype: float
        """
        return self.ds.getStickAxis(self.port, 5)
    
    def getRightPressed(self):
        """Determines if the right stick is pressed
        
        :returns: True if pressed, False otherwise
        :rtype: bool
        """
        return self.ds.getStickButton(self.port, 9)
    
    def getButtonA(self):
        """Gets whether the A button is pressed
        
        :returns: True if pressed, False otherwise
        :rtype: bool
        """
        return self.ds.getStickButton(self.port, 0)
    
    def getButtonB(self):
        """Gets whether the B button is pressed
        
        :returns: True if pressed, False otherwise
        :rtype: bool
        """
        return self.ds.getStickButton(self.port, 1)
    
    def getButtonX(self):
        """Gets whether the X button is pressed
        
        :returns: True if pressed, False otherwise
        :rtype: bool
        """
        return self.ds.getStickButton(self.port, 2)
    
    def getButtonY(self):
        """Gets whether the X button is pressed
        
        :returns: True if pressed, False otherwise
        :rtype: bool
        """
        return self.ds.getStickButton(self.port, 3)
    
    def getStart(self):
        """Gets whether the Start button is pressed
        
        :returns: True if pressed, False otherwise
        :rtype: bool
        """
        return self.ds.getStickButton(self.port, 7)
    
    def getBack(self):
        """Gets whether the Back button is pressed
        
        :returns: True if pressed, False otherwise
        :rtype: bool
        """
        return self.ds.getStickButton(self.port, 6)
    
    def getLeftBumper(self):
        """Gets whether the left bumper is pressed
        
        :returns: True if pressed, False otherwise
        :rtype: bool
        """
        return self.ds.getStickButton(self.port, 4)
    
    def getRightBumper(self):
        """Gets whether the right bumper is pressed
        
        :returns: True if pressed, False otherwise
        :rtype: bool
        """
        return self.ds.getStickButton(self.port, 5)
    
    def getLeftTrigger(self):
        """Gets whether the left trigger is pressed
        
        :returns: True if pressed, False otherwise
        :rtype: bool
        """
        return self.ds.getStickAxis(self.port, 2) > 0
    
    def getRightTrigger(self):
        """Gets whether the right trigger is pressed
        
        :returns: True if pressed, False otherwise
        :rtype: bool
        """
        return self.ds.getStickAxis(self.port, 3) > 0



#----------------------------------------------------------------------------
# Copyright (c) Kauai Labs 2015. All Rights Reserved.
#
# Created in support of Team 2465 (Kauaibots).  Go Purple Wave!
#
# Open Source Software - may be modified and shared by FRC teams. Any
# modifications to this code must be accompanied by the \License.txt file
# in the root directory of the project
#----------------------------------------------------------------------------

from .imuprotocol import IMUProtocol

class AHRSProtocol(IMUProtocol):
    
    
    # NAVX_CAL_STATUS 

    NAVX_CAL_STATUS_IMU_CAL_STATE_MASK =          0x03
    NAVX_CAL_STATUS_IMU_CAL_INPROGRESS =          0x00
    NAVX_CAL_STATUS_IMU_CAL_ACCUMULATE =          0x01
    NAVX_CAL_STATUS_IMU_CAL_COMPLETE =            0x02

    NAVX_CAL_STATUS_MAG_CAL_COMPLETE =            0x04
    NAVX_CAL_STATUS_BARO_CAL_COMPLETE =           0x08

    # NAVX_SELFTEST_STATUS 

    NAVX_SELFTEST_STATUS_COMPLETE = 0x80

    NAVX_SELFTEST_RESULT_GYRO_PASSED =            0x01
    NAVX_SELFTEST_RESULT_ACCEL_PASSED =           0x02
    NAVX_SELFTEST_RESULT_MAG_PASSED =             0x04
    NAVX_SELFTEST_RESULT_BARO_PASSED =            0x08

    # NAVX_OP_STATUS 

    NAVX_OP_STATUS_INITIALIZING =                 0x00
    NAVX_OP_STATUS_SELFTEST_IN_PROGRESS =         0x01
    NAVX_OP_STATUS_ERROR =                        0x02
    NAVX_OP_STATUS_IMU_AUTOCAL_IN_PROGRESS =      0x03
    NAVX_OP_STATUS_NORMAL =                       0x04

    # NAVX_SENSOR_STATUS 
    NAVX_SENSOR_STATUS_MOVING =                   0x01
    NAVX_SENSOR_STATUS_YAW_STABLE =               0x02
    NAVX_SENSOR_STATUS_MAG_DISTURBANCE =          0x04
    NAVX_SENSOR_STATUS_ALTITUDE_VALID =           0x08
    NAVX_SENSOR_STATUS_SEALEVEL_PRESS_SET =       0x10
    NAVX_SENSOR_STATUS_FUSED_HEADING_VALID =      0x20
    
    # NAVX_REG_CAPABILITY_FLAGS (Aligned w/NAV6 Flags, see IMUProtocol.h) 

    NAVX_CAPABILITY_FLAG_OMNIMOUNT =             0x0004
    NAVX_CAPABILITY_FLAG_OMNIMOUNT_CONFIG_MASK = 0x0038
    NAVX_CAPABILITY_FLAG_VEL_AND_DISP =          0x0040
    NAVX_CAPABILITY_FLAG_YAW_RESET =             0x0080

    # NAVX_OMNIMOUNT_CONFIG 

    OMNIMOUNT_DEFAULT =                       0 # Same as Y_Z_UP 
    OMNIMOUNT_YAW_X_UP =                      1
    OMNIMOUNT_YAW_X_DOWN =                    2
    OMNIMOUNT_YAW_Y_UP =                      3
    OMNIMOUNT_YAW_Y_DOWN =                    4
    OMNIMOUNT_YAW_Z_UP =                      5
    OMNIMOUNT_YAW_Z_DOWN =                    6

    # NAVX_INTEGRATION_CTL 

    NAVX_INTEGRATION_CTL_RESET_VEL_X =        0x01
    NAVX_INTEGRATION_CTL_RESET_VEL_Y =        0x02
    NAVX_INTEGRATION_CTL_RESET_VEL_Z =        0x04
    NAVX_INTEGRATION_CTL_RESET_DISP_X =       0x08
    NAVX_INTEGRATION_CTL_RESET_DISP_Y =       0x10
    NAVX_INTEGRATION_CTL_RESET_DISP_Z =       0x20
    NAVX_INTEGRATION_CTL_RESET_YAW =          0x80
    
    class AHRS_TUNING_VAR_ID:
        UNSPECIFIED =                  0
        MOTION_THRESHOLD =             1 # In G 
        YAW_STABLE_THRESHOLD =         2 # In Degrees 
        MAG_DISTURBANCE_THRESHOLD =    3 # Ratio 
        SEA_LEVEL_PRESSURE =           4 # Millibars 
    
    
    class AHRS_DATA_TYPE:
        TUNING_VARIABLE =  0
        MAG_CALIBRATION =  1
        BOARD_IDENTITY =   2

    class AHRS_DATA_ACTION:
        DATA_GET = 0
        DATA_SET = 1

    BINARY_PACKET_INDICATOR_CHAR = '#'

    # AHRS Protocol encodes certain data in binary format, unlike the IMU  
    # protocol, which encodes all data in ASCII characters.  Thus, the     
    # packet start and message termination sequences may occur within the  
    # message content itself.  To support the binary format, the binary    
    # message has this format:                                             
    #                                                                      
    # [start][binary indicator][len][msgid]<MESSAGE>[checksum][terminator] 
    #                                                                      
    # (The binary indicator and len are not present in the ASCII protocol) 
    #                                                                      
    # The [len] does not include the length of the start and binary        
    # indicator characters, but does include all other message items,      
    # including the checksum and terminator sequence.                      
 
    MSGID_AHRS_UPDATE = 'a'
    AHRS_UPDATE_YAW_VALUE_INDEX = 4  # Degrees.  Signed Hundredths 
    AHRS_UPDATE_PITCH_VALUE_INDEX = 6  # Degrees.  Signed Hundredeths 
    AHRS_UPDATE_ROLL_VALUE_INDEX = 8  # Degrees.  Signed Hundredths 
    AHRS_UPDATE_HEADING_VALUE_INDEX = 10  # Degrees.  Unsigned Hundredths 
    AHRS_UPDATE_ALTITUDE_VALUE_INDEX = 12 # Meters.   Signed 16:16 
    AHRS_UPDATE_FUSED_HEADING_VALUE_INDEX = 16 # Degrees.  Unsigned Hundredths 
    AHRS_UPDATE_LINEAR_ACCEL_X_VALUE_INDEX = 18 # Inst. G.  Signed Thousandths 
    AHRS_UPDATE_LINEAR_ACCEL_Y_VALUE_INDEX = 20 # Inst. G.  Signed Thousandths 
    AHRS_UPDATE_LINEAR_ACCEL_Z_VALUE_INDEX = 22 # Inst. G.  Signed Thousandths 
    AHRS_UPDATE_CAL_MAG_X_VALUE_INDEX = 24 # Int16 (Device Units) 
    AHRS_UPDATE_CAL_MAG_Y_VALUE_INDEX = 26 # Int16 (Device Units) 
    AHRS_UPDATE_CAL_MAG_Z_VALUE_INDEX = 28 # Int16 (Device Units) 
    AHRS_UPDATE_CAL_MAG_NORM_RATIO_VALUE_INDEX = 30 # Ratio.  Unsigned Hundredths 
    AHRS_UPDATE_CAL_MAG_SCALAR_VALUE_INDEX = 32 # Coefficient. Signed 16:16 
    AHRS_UPDATE_MPU_TEMP_VAUE_INDEX = 36 # Centigrade.  Signed Hundredths 
    AHRS_UPDATE_RAW_MAG_X_VALUE_INDEX = 38 # INT16 (Device Units 
    AHRS_UPDATE_RAW_MAG_Y_VALUE_INDEX = 40 # INT16 (Device Units 
    AHRS_UPDATE_RAW_MAG_Z_VALUE_INDEX = 42 # INT16 (Device Units 
    AHRS_UPDATE_QUAT_W_VALUE_INDEX = 44 # INT16 
    AHRS_UPDATE_QUAT_X_VALUE_INDEX = 46 # INT16 
    AHRS_UPDATE_QUAT_Y_VALUE_INDEX = 48 # INT16 
    AHRS_UPDATE_QUAT_Z_VALUE_INDEX = 50 # INT16     
    AHRS_UPDATE_BARO_PRESSURE_VALUE_INDEX = 52 # millibar.  Signed 16:16 
    AHRS_UPDATE_BARO_TEMP_VAUE_INDEX = 56 # Centigrade.  Signed  Hundredths 
    AHRS_UPDATE_OPSTATUS_VALUE_INDEX = 58 # NAVX_OP_STATUS_XXX 
    AHRS_UPDATE_SENSOR_STATUS_VALUE_INDEX = 59 # NAVX_SENSOR_STATUS_XXX 
    AHRS_UPDATE_CAL_STATUS_VALUE_INDEX    = 60 # NAVX_CAL_STATUS_XXX 
    AHRS_UPDATE_SELFTEST_STATUS_VALUE_INDEX = 61 # NAVX_SELFTEST_STATUS_XXX 
    AHRS_UPDATE_MESSAGE_CHECKSUM_INDEX = 62
    AHRS_UPDATE_MESSAGE_TERMINATOR_INDEX = 64
    AHRS_UPDATE_MESSAGE_LENGTH = 66

    # AHRSAndPositioning Update Packet (similar to AHRS, but removes magnetometer and adds velocity/displacement) 

    MSGID_AHRSPOS_UPDATE =             'p'
    AHRSPOS_UPDATE_YAW_VALUE_INDEX =           4 # Degrees.  Signed Hundredths 
    AHRSPOS_UPDATE_PITCH_VALUE_INDEX =         6 # Degrees.  Signed Hundredeths 
    AHRSPOS_UPDATE_ROLL_VALUE_INDEX =          8 # Degrees.  Signed Hundredths 
    AHRSPOS_UPDATE_HEADING_VALUE_INDEX =       10 # Degrees.  Unsigned Hundredths 
    AHRSPOS_UPDATE_ALTITUDE_VALUE_INDEX =      12 # Meters.   Signed 16:16 
    AHRSPOS_UPDATE_FUSED_HEADING_VALUE_INDEX = 16 # Degrees.  Unsigned Hundredths 
    AHRSPOS_UPDATE_LINEAR_ACCEL_X_VALUE_INDEX = 18 # Inst. G.  Signed Thousandths 
    AHRSPOS_UPDATE_LINEAR_ACCEL_Y_VALUE_INDEX = 20 # Inst. G.  Signed Thousandths 
    AHRSPOS_UPDATE_LINEAR_ACCEL_Z_VALUE_INDEX = 22 # Inst. G.  Signed Thousandths 
    AHRSPOS_UPDATE_VEL_X_VALUE_INDEX =         24 # Signed 16:16, in meters/sec 
    AHRSPOS_UPDATE_VEL_Y_VALUE_INDEX =         28 # Signed 16:16, in meters/sec 
    AHRSPOS_UPDATE_VEL_Z_VALUE_INDEX =         32 # Signed 16:16, in meters/sec 
    AHRSPOS_UPDATE_DISP_X_VALUE_INDEX =        36 # Signed 16:16, in meters 
    AHRSPOS_UPDATE_DISP_Y_VALUE_INDEX =        40 # Signed 16:16, in meters 
    AHRSPOS_UPDATE_DISP_Z_VALUE_INDEX =        44 # Signed 16:16, in meters 
    AHRSPOS_UPDATE_QUAT_W_VALUE_INDEX =        48 # INT16 
    AHRSPOS_UPDATE_QUAT_X_VALUE_INDEX =        50 # INT16 
    AHRSPOS_UPDATE_QUAT_Y_VALUE_INDEX =        52 # INT16 
    AHRSPOS_UPDATE_QUAT_Z_VALUE_INDEX =        54 # INT16 
    AHRSPOS_UPDATE_MPU_TEMP_VAUE_INDEX =       56 # Centigrade.  Signed Hundredths 
    AHRSPOS_UPDATE_OPSTATUS_VALUE_INDEX =      58 # NAVX_OP_STATUS_XXX 
    AHRSPOS_UPDATE_SENSOR_STATUS_VALUE_INDEX = 59 # NAVX_SENSOR_STATUS_XXX 
    AHRSPOS_UPDATE_CAL_STATUS_VALUE_INDEX =    60 # NAVX_CAL_STATUS_XXX 
    AHRSPOS_UPDATE_SELFTEST_STATUS_VALUE_INDEX    = 61 # NAVX_SELFTEST_STATUS_XXX 
    AHRSPOS_UPDATE_MESSAGE_CHECKSUM_INDEX =    62
    AHRSPOS_UPDATE_MESSAGE_TERMINATOR_INDEX =  64
    AHRSPOS_UPDATE_MESSAGE_LENGTH =            66
    
    # Data Get Request:  Tuning Variable, Mag Cal, Board Identity (Response message depends upon request type)
    MSGID_DATA_REQUEST =               'D'
    DATA_REQUEST_DATATYPE_VALUE_INDEX =        4
    DATA_REQUEST_VARIABLEID_VALUE_INDEX =      5
    DATA_REQUEST_CHECKSUM_INDEX =              6
    DATA_REQUEST_TERMINATOR_INDEX =            8
    DATA_REQUEST_MESSAGE_LENGTH =              10
    
    # Data Set Response Packet
    MSGID_DATA_SET_RESPONSE =          'v'
    DATA_SET_RESPONSE_DATATYPE_VALUE_INDEX =   4
    DATA_SET_RESPONSE_VARID_VALUE_INDEX =      5
    DATA_SET_RESPONSE_STATUS_VALUE_INDEX =     6
    DATA_SET_RESPONSE_MESSAGE_CHECKSUM_INDEX = 7
    DATA_SET_RESPONSE_MESSAGE_TERMINATOR_INDEX = 9
    DATA_SET_RESPONSE_MESSAGE_LENGTH =         11

    # Integration Control Command Packet 
    MSGID_INTEGRATION_CONTROL_CMD =    'I'
    INTEGRATION_CONTROL_CMD_ACTION_INDEX =     4
    INTEGRATION_CONTROL_CMD_PARAMETER_INDEX =  5
    INTEGRATION_CONTROL_CMD_MESSAGE_CHECKSUM_INDEX    = 9
    INTEGRATION_CONTROL_CMD_MESSAGE_TERMINATOR_INDEX = 11
    INTEGRATION_CONTROL_CMD_MESSAGE_LENGTH    =   13

    # Integration Control Response Packet 
    MSGID_INTEGRATION_CONTROL_RESP =   'i'
    INTEGRATION_CONTROL_RESP_ACTION_INDEX =    4
    INTEGRATION_CONTROL_RESP_PARAMETER_INDEX = 5
    INTEGRATION_CONTROL_RESP_MESSAGE_CHECKSUM_INDEX = 9
    INTEGRATION_CONTROL_RESP_MESSAGE_TERMINATOR_INDEX = 11
    INTEGRATION_CONTROL_RESP_MESSAGE_LENGTH =  13
    
    # Magnetometer Calibration Packet - e.g., !m[x_bias][y_bias][z_bias][m1,1 ... m3,3][cr][lf]
    MSGID_MAG_CAL_CMD =                'M'
    MAG_CAL_DATA_ACTION_VALUE_INDEX =          4
    MAG_X_BIAS_VALUE_INDEX =                   5 # signed short 
    MAG_Y_BIAS_VALUE_INDEX =                   7
    MAG_Z_BIAS_VALUE_INDEX =                   9
    MAG_XFORM_1_1_VALUE_INDEX =                11 # signed 16:16 
    MAG_XFORM_1_2_VALUE_INDEX =                15
    MAG_XFORM_1_3_VALUE_INDEX =                19
    MAG_XFORM_2_1_VALUE_INDEX =                23
    MAG_XFORM_2_2_VALUE_INDEX =                25
    MAG_XFORM_2_3_VALUE_INDEX =                31
    MAG_XFORM_3_1_VALUE_INDEX =                35
    MAG_XFORM_3_2_VALUE_INDEX =                39
    MAG_XFORM_3_3_VALUE_INDEX =                43
    MAG_CAL_EARTH_MAG_FIELD_NORM_VALUE_INDEX = 47
    MAG_CAL_CMD_MESSAGE_CHECKSUM_INDEX  =      51
    MAG_CAL_CMD_MESSAGE_TERMINATOR_INDEX =     53
    MAG_CAL_CMD_MESSAGE_LENGTH =               55

    # Tuning Variable Packet
    MSGID_FUSION_TUNING_CMD  =         'T'
    FUSION_TUNING_DATA_ACTION_VALUE_INDEX =    4
    FUSION_TUNING_CMD_VAR_ID_VALUE_INDEX =     5
    FUSION_TUNING_CMD_VAR_VALUE_INDEX =        6
    FUSION_TUNING_CMD_MESSAGE_CHECKSUM_INDEX = 10
    FUSION_TUNING_CMD_MESSAGE_TERMINATOR_INDEX = 12
    FUSION_TUNING_CMD_MESSAGE_LENGTH =         14

    # Board Identity Response Packet- e.g., !c[type][hw_rev][fw_major][fw_minor][unique_id[12]]
    MSGID_BOARD_IDENTITY_RESPONSE =    'i'
    BOARD_IDENTITY_BOARDTYPE_VALUE_INDEX =     4
    BOARD_IDENTITY_HWREV_VALUE_INDEX =         5
    BOARD_IDENTITY_FW_VER_MAJOR =              6
    BOARD_IDENTITY_FW_VER_MINOR =              7
    BOARD_IDENTITY_FW_VER_REVISION_VALUE_INDEX = 8
    BOARD_IDENTITY_UNIQUE_ID_0 =               10
    BOARD_IDENTITY_UNIQUE_ID_1 =               11
    BOARD_IDENTITY_UNIQUE_ID_2 =               12
    BOARD_IDENTITY_UNIQUE_ID_3 =               13
    BOARD_IDENTITY_UNIQUE_ID_4 =               14
    BOARD_IDENTITY_UNIQUE_ID_5 =               15
    BOARD_IDENTITY_UNIQUE_ID_6 =               16
    BOARD_IDENTITY_UNIQUE_ID_7 =               17
    BOARD_IDENTITY_UNIQUE_ID_8 =               18
    BOARD_IDENTITY_UNIQUE_ID_9 =               19
    BOARD_IDENTITY_UNIQUE_ID_10 =              20
    BOARD_IDENTITY_UNIQUE_ID_11 =              21
    BOARD_IDENTITY_RESPONSE_CHECKSUM_INDEX =   22
    BOARD_IDENTITY_RESPONSE_TERMINATOR_INDEX = 24
    BOARD_IDENTITY_RESPONSE_MESSAGE_LENGTH =   26

    MAX_BINARY_MESSAGE_LENGTH = AHRSPOS_UPDATE_MESSAGE_LENGTH

#     class AHRSUpdate:
#         yaw = 0.0
#         pitch = 0.0
#         roll = 0.0
#         compass_heading = 0.0
#         altitude = 0.0
#         fused_heading = 0.0
#         world_linear_accel_x = 0.0
#         world_linear_accel_y = 0.0
#         world_linear_accel_z = 0.0
#         cal_mag_x = 0
#         cal_mag_y = 0
#         cal_mag_z = 0
#         mag_field_norm_ratio = 0.0
#         mag_field_norm_scalar = 0.0
#         mpu_temp_c = 0.0
#         raw_mag_x = 0
#         raw_mag_y = 0
#         raw_mag_z = 0
#         quaternionW = 0
#         quaternionX = 0
#         quaternionY = 0
#         quaternionZ = 0
#         barometric_pressure = 0.0
#         baro_temp = 0.0
#         op_status = 0
#         sensor_status = 0
#         cal_status = 0
#         selftest_status = 0
        
    class AHRSPosUpdate:
        yaw = 0.0
        pitch = 0.0
        roll = 0.0
        compass_heading = 0.0
        altitude = 0.0
        fused_heading = 0.0
        world_linear_accel_x = 0.0
        world_linear_accel_y = 0.0
        world_linear_accel_z = 0.0
        vel_x = 0.0
        vel_y = 0.0
        vel_z = 0.0
        disp_x = 0.0
        disp_y = 0.0
        disp_z = 0.0
        mpu_temp = 0.0
        quaternionW = 0
        quaternionX = 0
        quaternionY = 0
        quaternionZ = 0
        barometric_pressure = 0.0
        baro_temp = 0.0
        op_status = 0
        sensor_status = 0
        cal_status = 0
        selftest_status = 0

    @staticmethod
    def decodeBinaryUint16(data, offset):
        return int.from_bytes(data[offset:offset+2], 'little', signed=False)
    
    @staticmethod
    def encodeBinaryUint16(i, data, offset):
        data[offset:offset+2] = i.to_bytes(2, 'little', signed=False)
    
    @staticmethod
    def decodeBinaryInt16(data, offset):
        return int.from_bytes(data[offset:offset+2], 'little', signed=True)
    
    @staticmethod
    def encodeBinaryInt16(i, data, offset):
        data[offset:offset+2] = i.to_bytes(2, 'little', signed=True)

    @staticmethod
    def decodeProtocol1616Float(data, offset):
        return int.from_bytes(data[offset:offset+4], 'little', signed=True)/65536.0
    
    @staticmethod
    def encodeProtocol1616Float(f, data, offset):
        data[offset:offset+4] = int(f*65536).to_bytes(4, 'little', signed=True)
    
    @staticmethod
    def decodeProtocolSignedHundredthsFloat(data, offset):
        return int.from_bytes(data[offset:offset+2], 'little', signed=True)/100.0
    
    @staticmethod
    def encodeProtocolSignedHundredthsFloat(f, data, offset):
        data[offset:offset+2] = int(f*100).to_bytes(2, 'little', signed=True)
    
    @staticmethod
    def decodeProtocolSignedThousandthsFloat(data, offset):
        return int.from_bytes(data[offset:offset+2], 'little', signed=True)/1000.0
    
    @staticmethod
    def encodeProtocolSignedThousandthsFloat(f, data, offset):
        data[offset:offset+2] = int(f*1000).to_bytes(2, 'little', signed=True)
    
    @staticmethod
    def decodeProtocolUnsignedHundredthsFloat(data, offset):
        return int.from_bytes(data[offset:offset+2], 'little', signed=False)/100.0
    
    @staticmethod
    def encodeProtocolUnsignedHundredthsFloat(f, data, offset):
        data[offset:offset+2] = int(f*100).to_bytes(2, 'little', signed=False)





class MagicComponent:
    """
        To automagically retrieve variables defined in your base robot
        object, you can add the following::
        
            class MyComponent:
            
                # other variables 'imported' automatically from MagicRobot
                elevator_motor = Talon
                other_component = MyOtherComponent
                
                ...
                
                def execute(self):
                
                    # This will be automatically set to the Talon
                    # instance created in robot.py
                    self.elevator_motor.set(self.value)
                
                
        What this says is "find the variable in the robot class called
        'elevator_motor', which is a Talon". If the name and type match,
        then the variable will automatically be injected into your
        component when it is created.
    
        .. note:: You don't need to inherit from ``MagicComponent``, it is only
                  provided for documentation's sake
    """
    
    def setup(self):
        """
            This function is called after ``createObjects`` has been called in
            the main robot class, and after all components have been created
            
            The setup function is optional and components do not have to define
            one. ``setup()`` functions are called in order of component definition
            in the main robot class.
            
            .. note:: For technical reasons, variables imported from
                      MagicRobot are not initialized when your component's
                      constructor is called. However, they will be initialized
                      by the time this function is called.   
        """
    
    def on_enabled(self):
        """
            Called when the robot enters autonomous or teleoperated mode. This
            function should initialize your component to a "safe" state so
            that unexpected things don't happen when enabling the robot.
            
            .. note:: You'll note that there isn't a separate initialization
                      function for autonomous and teleoperated modes. This is
                      intentional, as they should be the same.
        """
        
        
    def on_disabled(self):
        """
            Called when the robot leaves autonomous or teleoperated
        """
    
    
    def execute(self):
        """
            This function is called at the end of the control loop
        """
    


#----------------------------------------------------------------------------
# Copyright (c) Kauai Labs 2015. All Rights Reserved.
#
# Created in support of Team 2465 (Kauaibots).  Go Purple Wave!
#
# Open Source Software - may be modified and shared by FRC teams. Any
# modifications to this code must be accompanied by the \License.txt file
# in the root directory of the project
#----------------------------------------------------------------------------

import enum

__all__ = ['getNavxDigitalChannel',
           'getNavxPWMChannel',
           'getNavxAnalogInChannel',
           'getNavxAnalogOutChannel',
           'getChannelFromPin',
           'PinType']

MAX_NAVX_MXP_DIGIO_PIN_NUMBER      = 9
MAX_NAVX_MXP_ANALOGIN_PIN_NUMBER   = 3
MAX_NAVX_MXP_ANALOGOUT_PIN_NUMBER  = 1
NUM_ROBORIO_ONBOARD_DIGIO_PINS     = 10
NUM_ROBORIO_ONBOARD_PWM_PINS       = 10
NUM_ROBORIO_ONBOARD_ANALOGIN_PINS  = 4
NUM_ROBORIO_ONBOARD_ANALOGOUT_PINS = 2


def getNavxDigitalChannel(pin):
    """Get the WPILib channel number for a digital input on a NavX MXP
    
    :param pin: The NavX pin number (0-9)
    :type pin: int
    
    :returns: RoboRIO channel number
    """
    if pin < 0 or pin > MAX_NAVX_MXP_DIGIO_PIN_NUMBER:
        raise ValueError("Invalid navX MXP Digital I/O pin number")
    
    return pin + NUM_ROBORIO_ONBOARD_DIGIO_PINS + (4 if pin > 3 else 0);
    
def getNavxPWMChannel(pin):
    """Get the WPILib channel number for a PWM on a NavX MXP
    
    :param pin: The NavX pin number (0-9)
    :type pin: int
    
    :returns: RoboRIO channel number
    """
    if pin < 0 or pin > MAX_NAVX_MXP_DIGIO_PIN_NUMBER:
        raise ValueError("Invalid navX MXP PWM pin number")
    
    return pin + NUM_ROBORIO_ONBOARD_PWM_PINS
    
def getNavxAnalogInChannel(pin):
    """Get the WPILib channel number for an Analog Input on a NavX MXP
    
    :param pin: The NavX pin number (0-3)
    :type pin: int
    
    :returns: RoboRIO channel number
    """
    if pin < 0 or pin > MAX_NAVX_MXP_ANALOGIN_PIN_NUMBER:
        raise ValueError("Invalid navX MXP Analog In pin number")
    
    return pin + NUM_ROBORIO_ONBOARD_ANALOGIN_PINS
    
def getNavxAnalogOutChannel(pin):
    """Get the WPILib channel number for an Analog Output on a NavX MXP
    
    :param pin: The NavX pin number (0-1)
    :type pin: int
    
    :returns: RoboRIO channel number
    """
    if pin < 0 or pin > MAX_NAVX_MXP_ANALOGOUT_PIN_NUMBER:
        raise ValueError("Invalid navX MXP Analog Out pin number")
    
    return pin


class PinType(enum.Enum):
    DigitalIO = 1
    PWM = 2
    AnalogIn = 3
    AnalogOut = 4


def getChannelFromPin(pinType, pin):
    """Converts from a NavX MXP pin type and number to the corresponding
    RoboRIO channel number that is used by WPILib.
    
    :param pinType: The type of pin to convert 
    :type pinType: :class:`PinType`
    :param pin: 
    :type pin: int
    
    :returns: RoboRIO channel number
    """
    
    if pinType == PinType.DigitalIO:
        return getNavxDigitalChannel(pin)
    
    if pinType == PinType.PWM:
        return getNavxPWMChannel(pin)
    
    if pinType == PinType.AnalogIn:
        return getNavxAnalogInChannel(pin)
    
    if pinType == PinType.AnalogOut:
        return getNavxAnalogOutChannel(pin)
    
    raise ValueError("Invalid pin type specified")




import functools
import inspect
import networktables

from robotpy_ext.misc.orderedclass import OrderedClass

from .magic_tunable import tunable

from wpilib import Timer

class IllegalCallError(Exception):
    pass

class NoFirstStateError(Exception):
    pass

class MultipleFirstStatesError(Exception):
    pass


def _create_wrapper(f, first, must_finish):
    
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        raise IllegalCallError("Do not call states directly, use begin/next_state instead")
    
    # store state variables here
    wrapper.name = f.__name__
    wrapper.description = f.__doc__
    wrapper.first = first
    wrapper.must_finish = must_finish
    
    # inspect the args, provide a correct call implementation
    args, varargs, keywords, _ = inspect.getargspec(f)
    
    if keywords is not None:
        raise ValueError("Cannot use keyword arguments for function %s" % wrapper.name)
    
    if varargs is not None:
        raise ValueError("Cannot use *args arguments for function %s" % wrapper.name)
    
    if args[0] != 'self':
        raise ValueError("First argument must be 'self'")
    
    # TODO: there has to be a better way to do this. oh well. it only runs once.
    
    if len(args) > 4:
        raise ValueError("Too many parameters for %s" % wrapper.name)
    
    wrapper_creator = 'w = lambda self, tm, state_tm, initial_call: f(%s)' % ','.join(args)
    
    for arg in ['self', 'tm', 'state_tm', 'initial_call']:
        try:
            args.remove(arg)
        except ValueError:
            pass
    
    if len(args) != 0:
        raise ValueError("Invalid parameter names in %s: %s" % (wrapper.name, ','.join(args)))
    
    varlist = {'f': f}
    exec(wrapper_creator, varlist, varlist)
    
    wrapper.run = varlist['w']    
    return wrapper

class _StateData:
    def __init__(self, wrapper):
        self.name = wrapper.name
        self.duration_attr = '%s_duration' % self.name
        self.expires = 0xffffffff
        self.ran = False
        self.run = wrapper.run
        self.must_finish = wrapper.must_finish
        
        if hasattr(wrapper, 'next_state'):
            self.next_state = wrapper.next_state

def timed_state(f=None, *, duration=None, next_state=None, first=False, must_finish=False):
    '''
        If this decorator is applied to a function in an object that inherits
        from :class:`.StateMachine`, it indicates that the function
        is a state that will run for a set amount of time unless interrupted
        
        The decorated function can have the following arguments in any order:
        
        - ``tm`` - The number of seconds since the state machine has started
        - ``state_tm`` - The number of seconds since this state has been active
          (note: it may not start at zero!)
        - ``initial_call`` - Set to True when the state is initially called,
          False otherwise. If the state is switched to multiple times, this
          will be set to True at the start of each state execution.
        
        :param duration: The length of time to run the state before progressing
                         to the next state
        :type  duration: float
        :param next_state: The name of the next state. If not specified, then
                           this will be the last state executed if time expires
        :type  next_state: str
        :param first: If True, this state will be ran first
        :type  first: bool
        :param must_finish: If True, then this state will continue executing
                            even if ``engage()`` is not called. However,
                            if ``done()`` is called, execution will stop
                            regardless of whether this is set.
        :type  must_finish: bool
    '''
    
    if f is None:
        return functools.partial(timed_state, duration=duration, next_state=next_state, first=first, must_finish=must_finish)
    
    if duration is None:
        raise ValueError("timed_state functions must specify a duration")
    
    wrapper = _create_wrapper(f, first, must_finish)
    
    wrapper.next_state = next_state
    wrapper.duration = duration
    
    return wrapper

def state(f=None, *, first=False, must_finish=False):
    '''
        If this decorator is applied to a function in an object that inherits
        from :class:`.StateMachine`, it indicates that the function
        is a state. The state will continue to be executed until the
        ``next_state`` function is executed.
        
        The decorated function can have the following arguments in any order:
        
        - ``tm`` - The number of seconds since the state machine has started
        - ``state_tm`` - The number of seconds since this state has been active
          (note: it may not start at zero!)
        - ``initial_call`` - Set to True when the state is initially called,
          False otherwise. If the state is switched to multiple times, this
          will be set to True at the start of each state execution.
        
        :param first: If True, this state will be ran first
        :type  first: bool
        :param must_finish: If True, then this state will continue executing
                            even if ``engage()`` is not called. However,
                            if ``done()`` is called, execution will stop
                            regardless of whether this is set.
        :type  must_finish: bool
    '''
    
    if f is None:
        return functools.partial(state, first=first, must_finish=must_finish)
    
    return _create_wrapper(f, first, must_finish)



class StateMachine(metaclass=OrderedClass):
    '''
        This object is designed to be used to easily implement magicbot
        components that are basically a big state machine.
        
        You use this by defining a class that inherits from ``StateMachine``.
        To define each state, you use the :func:`timed_state` decorator on a
        function. When each state is run, the decorated function will be
        called. Decorated state functions can receive the following
        parameters:
        
        - ``tm`` - The number of seconds since autonomous has started
        - ``state_tm`` - The number of seconds since this state has been active
          (note: it may not start at zero!)
        - ``initial_call`` - Set to True when the state is initially called,
          False otherwise. If the state is switched to multiple times, this
          will be set to True at the start of each state.
          
        To be consistent with the magicbot philosophy, in order for the
        state machine to execute its states you must call the
        :func:`engage` function upon each execution of the main
        robot control loop. If you do not call this function, then
        execution will cease unless the current executing state is
        marked as ``must_finish``.
        
        When execution ceases, the :func:`done` function will be called
        unless execution was stopped by calling the ``done`` function.
        
        As a magicbot component, this contains an ``execute`` function that
        will be called on each control loop. All state execution occurs from
        within that function call. If you call other components from this
        component, you should ensure that your component occurs *before*
        the other components in your Robot class.
        
        Here's a very simple example of how you might implement a shooter
        automation component that moves a ball into a shooter when the
        shooter is ready::
        
            class ShooterAutomation:
            
                # Some other component
                shooter = Shooter
                ball_pusher = BallPusher
                
                def fire(self):
                    """This is called from the main loop"""
                    self.engage()
                    
                @state(first=True)
                def begin_firing(self):
                    self.shooter.enable()
                    if self.shooter.ready():
                        self.next_state('firing')
                    
                @timed_state(duration=1.0, must_finish=True)
                def firing(self):
                    """This state will continue executing even if engage isn't called"""
                    self.shooter.enable()
                    self.ball_pusher.push()
            
            ...
                
            class MyRobot(magicbot.MagicRobot):
            
                ...
                
                def teleopPeriodic(self):
                
                    if self.joystick.getTrigger():
                        self.shooter_automation.fire()
                        
        This object has a lot of really useful NetworkTables integration
        as well:
        
        - tunables are created in /components/NAME/state
          - state durations can be tuned here
          - The 'current state' is output as it happens
          - Descriptions and names of the states are here (for dashboard use)
          
        
        .. warning:: This object is not intended to be threadsafe and should not
                     be accessed from multiple threads
    '''
    
    VERBOSE_LOGGING = False
    
    #: NT variable that indicates which state will be executed next (though,
    #: does not guarantee that it will be executed). Will return an empty
    #: string if the state machine is not currently engaged.
    current_state = tunable('', subtable='state')
    
    def __new__(cls):
        # choose to use __new__ instead of __init__
        o = super().__new__(cls)
        o._build_states()
        return o
        
        # TODO: when this gets invoked, tunables need to be setup on
        # the object first
        
    def _build_states(self):
        has_first = False
    
        # problem: the user interface won't know which entries are the
        #          current variables being used by the robot. So, we setup
        #          an array with the names, and the dashboard uses that
        #          to determine the ordering too
        
        nt_names = []
        nt_desc = []
    
        states = {}
        cls = self.__class__
    
        #for each state function:
        for name in self.members:
            
            state = getattr(cls, name, None)
            if state is None or name.startswith('__') or not hasattr(state, 'first'):
                continue

            # is this the first state to execute?
            if state.first:
                if has_first:
                    raise MultipleFirstStatesError("Multiple states were specified as the first state!")
                
                self.__first = name
                has_first = True
            
            description = ''
            if state.description is not None:
                description = state.description
            
            state_data = _StateData(state)
            states[state.name] = state_data
            nt_names.append(state.name)
            nt_desc.append(description)
            
            # make the time tunable
            # -> this depends on tunables being bound after this function is called
            if hasattr(state, 'duration'):
                # don't create it twice (in case of mutliple instances)
                prop = getattr(cls, state_data.duration_attr, None)
                if prop is None:
                    prop = setattr(cls, state_data.duration_attr, tunable(state.duration, writeDefault=False, subtable='state'))
         
        if not has_first:
            raise NoFirstStateError("Starting state not defined! Use first=True on a state decorator")
        
        cls.state_names = tunable(nt_names, subtable='state')
        cls.state_descriptions = tunable(nt_desc, subtable='state')
        
        # Indicates that an external party wishes the state machine to execute
        self.__should_engage = False
        
        # Indicates that the state machine is currently executing
        self.__engaged = False
        
        # A dictionary of states
        self.__states = states
        
        # The currently executing state, or None if not executing
        self.__state = None
        
        
    @property
    def is_executing(self):
        ''':returns: True if the state machine is executing states'''
        #return self.__state is not None
        return self.__engaged
        
    def on_enable(self):
        '''
            magicbot component API: called when autonomous/teleop is enabled
        '''
        pass
    
    def on_disable(self):
        '''
            magicbot component API: called when autonomous/teleop is disabled
        '''
        self.done()
        
    def engage(self, initial_state=None, force=False):
        '''
            This signals that you want the state machine to execute its
            states.
            
            :param initial_state: If specified and execution is not currently
                                  occurring, start in this state instead of
                                  in the 'first' state
            :param force:         If True, will transition even if the state
                                  machine is currently active.
        '''
        self.__should_engage = True
    
        if force or self.__state is None:
            if initial_state:
                self.next_state(initial_state)
            else:
                self.next_state(self.__first)
    
    def next_state(self, name):
        '''Call this function to transition to the next state
        
        :param name: Name of the state to transition to
        
        .. note:: This should only be called from one of the state functions
        '''
        state = self.__states[name]
        state.ran = False
        self.current_state = state.name
        
        self.__state = state
        
    def next_state_now(self, name):
        '''Call this function to transition to the next state, and call the next
        state function immediately. Prefer to use :meth:`next_state` instead.
        
        :param name: Name of the state to transition to
        
        .. note:: This should only be called from one of the state functions
        '''
        self.next_state(name)
        # TODO: may want to do this differently?
        self.execute()
        
    def done(self):
        '''Call this function to end execution of the state machine.
        
        .. note:: If you wish to do something each time execution ceases,
                  override this function (but be sure to call
                  ``super().done()``!)
        '''
        if self.VERBOSE_LOGGING and self.__state is not None:
            self.logger.info("Stopped state machine execution")
        
        self.__state = None
        self.__engaged = False
        self.current_state = ''
        
        
    
    def execute(self):
        '''
            magicbot component API: This is called on each iteration of the
            control loop. Most of the time, you will not want to override
            this function.
        '''
        
        now = Timer.getFPGATimestamp()
        
        if not self.__engaged:
            if self.__should_engage:
                self.__start = now
                self.__engaged = True
            else:
                return
        
        tm = now - self.__start
        state = self.__state
        
        # we adjust this so that if we have states chained together,
        # then the total time it runs is the amount of time of the
        # states. Otherwise, the time drifts.
        new_state_start = tm
        
        if state is None:
            self.done()
        
        # determine if the time has passed to execute the next state
        # -> intentionally comes first, 
        if state is not None and state.expires < tm:
            if state.next_state is None:
                state = None
            else:
                self.next_state(state.next_state)
                new_state_start = state.expires
                state = self.__state
        
        if state is None or (not self.__should_engage and not state.must_finish):
            self.done()
        else:
            # is this the first time this was executed?
            initial_call = not state.ran
            if initial_call:
                state.ran = True
                state.start_time = new_state_start
                state.expires = new_state_start + getattr(self, state.duration_attr, 0xffffffff)
                
                if self.VERBOSE_LOGGING:
                    self.logger.info("%.3fs: Entering state: %s", tm, state.name)
            
            # execute the state function, passing it the arguments
            state.run(self, tm, tm - state.start_time, initial_call)
        
        # Reset this each time
        self.__should_engage = False

    
class AutonomousStateMachine(StateMachine):
    '''
        This is a specialized version of the StateMachine that is designed
        to be used as an autonomous mode. There are a few key differences:
        
        - The :func:`.engage` function is always called, so the state machine
          will always run to completion unless done() is called
        - VERBOSE_LOGGING is set to True, so a log message will be printed out upon
          each state transition
        
    '''
    
    VERBOSE_LOGGING = True
    
    def on_enable(self):
        super().on_enable()
        self.__engaged = True
    
    def on_iteration(self, tm):
        # TODO, remove the on_iteration function in 2017? 
        
        # Only engage the state machine until its execution finishes, otherwise
        # it will just keep repeating
        #
        # This is because if you keep calling engage(), the state machine will
        # loop. I'm tempted to change that, but I think it would lead to unexpected
        # side effects. Will have to contemplate this...
        
        if self.__engaged:
            self.engage()
            self.execute()
            self.__engaged = self.is_executing



'''
    Taken from the python documentation, distributed under
    that same license
'''

import collections

class OrderedClass(type):
    '''
        Metaclass that stores class attributes in a member called
        'members'. If you subclass something that uses this metaclass,
        the base class members will be listed before the subclass
    '''

    @classmethod
    def __prepare__(metacls, name, bases, **kwds):
        return collections.OrderedDict()

    def __new__(cls, name, bases, namespace, **kwds):
        result = type.__new__(cls, name, bases, dict(namespace))
        members = getattr(result, 'members', ())
        result.members = members + tuple(namespace)
        return result



"""
These are a set of drivers for the XL-MaxSonar EZ series of sonar modules.
The devices have a few different ways of reading from them, and the these drivers attempt to cover
some of the methods
"""
import wpilib

from . import driver_base
from . import units


class MaxSonarEZPulseWidth(wpilib.SensorBase, driver_base.DriverBase):
    """
    This is a driver for the MaxSonar EZ series of sonar sensors, using the pulse-width output of the sensor.

    To use this driver, pin 2 on the sensor must be mapped to a dio pin.
    """

    verified = True

    def __init__(self, channel, output_units=units.inch):
        """Sonar sensor constructor

        :param channel: The digital input index which is wired to the pulse-width output pin (pin 2) on the sensor.
        :param output_units: The Unit instance specifying the format of value to return
        """

        #Save value
        self.output_units = output_units

        #Setup the counter
        self.counter = wpilib.Counter(channel)
        self.counter.setSemiPeriodMode(highSemiPeriod=True)

        #Call the parents
        super().__init__()

    def free(self):
        """Delete (free) the counter for the sonar sensor"""
        self.counter.free()

    def get(self):
        """Return the current sonar sensor reading, in the units specified from the constructor"""
        inches = self.counter.getPeriod()/0.000147
        return units.convert(units.inch, self.output_units, inches)


class MaxSonarEZAnalog(wpilib.SensorBase, driver_base.DriverBase):
    """
    This is a driver for the MaxSonar EZ series of sonar sensors, using the analog output of the sensor.

    To use this driver, pin 3 on the sensor must be mapped to an analog pin, and the sensor must be on a 5v supply.
    """

    #This code has actually never been run, so it is extra not-verified!
    verified = False

    def __init__(self, channel, output_units=units.inch):
        """Sonar sensor constructor

        :param channel: The analog input index which is wired to the analog output pin (pin 3) on the sensor.
        :param output_units: The Unit instance specifying the format of value to return
        """

        #Save value
        self.output_units = output_units

        #Setup the analog input
        self.analog = wpilib.AnalogInput(channel)
        
        #Call the parents
        super().__init__()

    def free(self):
        """Delete (free) the analog input for the sonar sensor"""
        self.analog.free()

    def get(self):
        """Return the current sonar sensor reading, in the units specified from the constructor"""
        centimeters = self.analog.getVoltage()/.0049
        return units.convert(units.centimeter, self.output_units, centimeters)




from wpilib.command import InstantCommand

class StopCommand(InstantCommand):
    '''
    When this command is run, it cancels the command it was passed, even if that
    command is part of a :class:`wpilib.CommandGroup`.
    '''


    def __init__(self, command):
        '''
            :param command: The :class:`wpilib.Command` to stop.
        '''
        super().__init__('Stop %s' % command)

        self.command = command


    def initialize(self):
        self.command._cancel()




from networktables.util import ChooserControl

autonomous_seconds = 15

def test_all_autonomous(control, fake_time, robot):
    '''
        This test runs all possible autonomous modes that can be selected
        by the autonomous switcher.
        
        This should work for most robots. If it doesn't work for yours, 
        and it's not a code issue with your robot, please file a bug on
        github.
    '''
    
    class AutonomousTester:
        
        def __init__(self):
            self.initialized = False
            self.init_time = None
            self.chooser = None
            
            self.state = 'auto'
            self.currentChoice = None
            self.until = None
        
        def initialize_chooser(self, tm):
            
            if self.chooser is None:
                self.chooser = ChooserControl('Autonomous Mode')
            
            self.choices = self.chooser.getChoices()
            if len(self.choices) == 0:
                return False
            
            self.state = 'disabled'
            self.currentChoice =  -1
            self.until = tm
            self.init_time = tm
            self.initialized = True
            return True
        
        def on_step(self, tm):
            
            if not self.initialized:
                if not self.initialize_chooser(tm):
                    assert tm < 10, "Robot didn't create a chooser within 10 seconds, probably an error"
                    return True
            
            if self.state == 'auto':
                if tm >= self.until:
                    self.until = tm + 1
                    self.state = 'disabled'
                    control.set_operator_control(enabled=False)
            
            elif self.state == 'disabled':
                if tm >= self.until:
                    
                    control.set_autonomous()
                    
                    self.state = 'auto'
                    self.until = tm + autonomous_seconds
                    self.currentChoice += 1
                    if self.currentChoice >= len(self.choices):
                        return False
                    
                    self.chooser.setSelected(self.choices[self.currentChoice])
            
            return True
    
    
    controller = control.run_test(AutonomousTester)
    
    # Make sure they ran for the correct amount of time
    assert int(fake_time.get()) == int(len(controller.choices)*(autonomous_seconds+1) + controller.init_time)




from . import AHRSProtocol
from . import IMURegisters

from hal_impl.data import hal_data

from hal_impl.i2c_helpers import I2CSimBase
from hal_impl.spi_helpers import SPISimBase

from ....misc.crc7 import crc7

class NavXSimBase:
    """
        This is a really quick hack to get the NavX working in simulation.
        
        If you wish to do more advanced things with this, please add the
        features and make a pull request. :)
    """
    
    def __init__(self):
        self.cap_flags = AHRSProtocol.NAVX_CAPABILITY_FLAG_YAW_RESET
    
    def _write(self, data):
        
        reg = data[0] & 0x7F
        
        if reg == IMURegisters.NAVX_REG_INTEGRATION_CTL:
            if data[0] == AHRSProtocol.NAVX_INTEGRATION_CTL_RESET_YAW:
                hal_data['robot'][self.angle_key] = 0
                
    
    def _read(self, data, count):
        # We only really support two calls, so cheat here
        
        # config buffer
        if count == 18:
            data[IMURegisters.NAVX_REG_WHOAMI] = IMURegisters.NAVX_MODEL_NAVX_MXP
            AHRSProtocol.encodeBinaryUint16(self.cap_flags, data, IMURegisters.NAVX_REG_CAPABILITY_FLAGS_L)
        
        # data
        elif count == 82:
            
            AHRSProtocol.encodeBinaryUint16(self.cap_flags, data, IMURegisters.NAVX_REG_CAPABILITY_FLAGS_L)
            
            # NavX returns angle in 180 to -180, angle key is continuous
            
            angle = hal_data['robot'].get(self.angle_key, 0)
            angle = ((angle + 180) % 360.0) - 180.0
            AHRSProtocol.encodeProtocolSignedHundredthsFloat(angle, data, IMURegisters.NAVX_REG_YAW_L-4)
            
        # no idea
        else:
            raise NotImplementedError("Oops")
        
class NavXI2CSim(NavXSimBase, I2CSimBase):
    
    def initializeI2C(self, port, status):
        status.value = 0
        
        # TODO: support other sim parameters
        self.angle_key = 'navxmxp_i2c_%d_angle' % port
        
    def writeI2C(self, port, deviceAddress, dataToSend, sendSize):
        self._write(dataToSend)
        return sendSize
    
    def readI2C(self, port, deviceAddress, buffer, count):
        self._read(buffer, count)
        return count

class NavXSPISim(NavXSimBase, SPISimBase):
    
    def initializeSPI(self, port, status):
        status.value = 0
        
        # TODO: support other sim parameters
        self.angle_key = 'navxmxp_spi_%d_angle' % port
    
    def writeSPI(self, port, dataToSend, sendSize):
        self._write(dataToSend)
        return sendSize
    
    def transactionSPI(self, port, dataToSend, dataReceived, size):
        self._read(dataReceived, size - 1)
        # TODO: maybe disable crc in sim
        dataReceived[-1] = crc7(dataReceived[:-1])
        return size




#----------------------------------------------------------------------------
# Copyright (c) Kauai Labs 2015. All Rights Reserved.
#
# Created in support of Team 2465 (Kauaibots).  Go Purple Wave!
#
# Open Source Software - may be modified and shared by FRC teams. Any
# modifications to this code must be accompanied by the \License.txt file
# in the root directory of the project
#----------------------------------------------------------------------------

from ._impl import AHRSProtocol, IMUProtocol, IMURegisters

from wpilib.timer import Timer

import logging
logger = logging.getLogger('navx')

__all__ = ['RegisterIO']

IO_TIMEOUT_SECONDS = 1.0



class _BoardId:
    type = 0
    hw_rev = 0
    fw_ver_major = 0
    fw_ver_minor = 0
    fw_revision = 0
    unique_id = [0]*12

class _BoardState:
    op_status = 0
    sensor_status = 0
    cal_status = 0
    selftest_status = 0
    capability_flags = 0
    update_rate_hz = 0
    accel_fsr_g = 0
    gyro_fsr_dps = 0

class RegisterIO:
    
    def __init__(self, io_provider, update_rate_hz, notify_sink, board_capabilities):
        """
            :param board_capabilities: must have the following callable attributes:
                _isOmniMountSupported, _isBoardYawResetSupported,
                _isDisplacementSupported
        
            :param notify_sink: must have the following callable attributes: 
                _setYawPitchRoll, _setAHRSData, _setAHRSPosData,
                _setRawData, _setBoardID, _setBoardState
        """
        
        self.io_provider = io_provider
        self.update_rate_hz = update_rate_hz
        self.board_capabilities = board_capabilities
        self.notify_sink = notify_sink
        
        self.raw_data_update = IMUProtocol.GyroUpdate()
        self.ahrspos_update = AHRSProtocol.AHRSPosUpdate()
        self.board_state = _BoardState()
        self.board_id = _BoardId()
        
        self.last_update_time = 0
        self.byte_count = 0
        self.update_count = 0
        self._stop = False
        
    def stop(self):
        self._stop = True
        
    def shutdown(self):
        self.io_provider.shutdown()
        
    def run(self):
        logger.info("NavX io thread starting")
        
        try:
            self.io_provider.init()
            
            # initial device configuration
            self.setUpdateRateHz(self.update_rate_hz)
            if not self.getConfiguration():
                logger.warn("-- Did not get configuration data")
            else:
                logger.info("-- Board is %s (rev %s)",
                            IMURegisters.model_type(self.board_id.type),
                            self.board_id.hw_rev)
                logger.info("-- Firmware %s.%s", self.board_id.fw_ver_major,
                                                 self.board_id.fw_ver_minor)
            
            log_error = True
            
            # IO Loop
            while not self._stop:
                if self.board_state.update_rate_hz != self.update_rate_hz:
                    self.setUpdateRateHz(self.update_rate_hz)
                
                try:
                    self.getCurrentData()
                except IOError:
                    if log_error:
                        logger.exception("Error getting data")
                        log_error = False
                else:
                    log_error = True
                    
                    
                Timer.delay(1.0/self.update_rate_hz)
        except Exception:
            logger.exception("Unhandled exception in NavX thread")
        finally:
            logger.info("NavX i/o thread exiting")
            
    def getConfiguration(self):
        success = False
        retry_count = 0
        
        while retry_count < 5 and not success:
            try:
                config = self.io_provider.read(IMURegisters.NAVX_REG_WHOAMI,
                                               IMURegisters.NAVX_REG_SENSOR_STATUS_H+1)
            except IOError as e:
                logger.warn("Error reading configuration data, retrying (%s)", e)
                success = False
                Timer.delay(0.5)
            else:
                board_id = self.board_id
                board_id.hw_rev                 = config[IMURegisters.NAVX_REG_HW_REV]
                board_id.fw_ver_major           = config[IMURegisters.NAVX_REG_FW_VER_MAJOR]
                board_id.fw_ver_minor           = config[IMURegisters.NAVX_REG_FW_VER_MINOR]
                board_id.type                   = config[IMURegisters.NAVX_REG_WHOAMI]
                self.notify_sink._setBoardID(board_id)
                
                board_state = self.board_state
                board_state.cal_status          = config[IMURegisters.NAVX_REG_CAL_STATUS]
                board_state.op_status           = config[IMURegisters.NAVX_REG_OP_STATUS]
                board_state.selftest_status     = config[IMURegisters.NAVX_REG_SELFTEST_STATUS]
                board_state.sensor_status       = AHRSProtocol.decodeBinaryUint16(config,IMURegisters.NAVX_REG_SENSOR_STATUS_L)
                board_state.gyro_fsr_dps        = AHRSProtocol.decodeBinaryUint16(config,IMURegisters.NAVX_REG_GYRO_FSR_DPS_L)
                board_state.accel_fsr_g         = config[IMURegisters.NAVX_REG_ACCEL_FSR_G]
                board_state.update_rate_hz      = config[IMURegisters.NAVX_REG_UPDATE_RATE_HZ]
                board_state.capability_flags    = AHRSProtocol.decodeBinaryUint16(config,IMURegisters.NAVX_REG_CAPABILITY_FLAGS_L)
                self.notify_sink._setBoardState(board_state)
                success = True
        
            retry_count += 1
        
        return success
    
    def getCurrentData(self):
        first_address = IMURegisters.NAVX_REG_UPDATE_RATE_HZ
        displacement_registers = self.board_capabilities._isDisplacementSupported()
        
        # If firmware supports displacement data, acquire it - otherwise implement
        # similar (but potentially less accurate) calculations on this processor.
        if displacement_registers:
            read_count = IMURegisters.NAVX_REG_LAST + 1 - first_address
        else:
            read_count = IMURegisters.NAVX_REG_QUAT_OFFSET_Z_H + 1 - first_address
        
        curr_data = self.io_provider.read(first_address, read_count)
    
        #timestamp_low = AHRSProtocol.decodeBinaryUint16(curr_data, IMURegisters.NAVX_REG_TIMESTAMP_L_L-first_address)
        #timestamp_high = AHRSProtocol.decodeBinaryUint16(curr_data, IMURegisters.NAVX_REG_TIMESTAMP_H_L-first_address)
        #sensor_timestamp            = (timestamp_high << 16) + timestamp_low
        ahrspos_update = self.ahrspos_update
        ahrspos_update.op_status       = curr_data[IMURegisters.NAVX_REG_OP_STATUS - first_address]
        ahrspos_update.selftest_status = curr_data[IMURegisters.NAVX_REG_SELFTEST_STATUS - first_address]
        ahrspos_update.cal_status      = curr_data[IMURegisters.NAVX_REG_CAL_STATUS]
        ahrspos_update.sensor_status   = curr_data[IMURegisters.NAVX_REG_SENSOR_STATUS_L - first_address]
        ahrspos_update.yaw             = AHRSProtocol.decodeProtocolSignedHundredthsFloat(curr_data, IMURegisters.NAVX_REG_YAW_L-first_address)
        ahrspos_update.pitch           = AHRSProtocol.decodeProtocolSignedHundredthsFloat(curr_data, IMURegisters.NAVX_REG_PITCH_L-first_address)
        ahrspos_update.roll            = AHRSProtocol.decodeProtocolSignedHundredthsFloat(curr_data, IMURegisters.NAVX_REG_ROLL_L-first_address)
        ahrspos_update.compass_heading = AHRSProtocol.decodeProtocolUnsignedHundredthsFloat(curr_data, IMURegisters.NAVX_REG_HEADING_L-first_address)
        ahrspos_update.mpu_temp_c      = AHRSProtocol.decodeProtocolSignedHundredthsFloat(curr_data, IMURegisters.NAVX_REG_MPU_TEMP_C_L - first_address)
        ahrspos_update.world_linear_accel_x  = AHRSProtocol.decodeProtocolSignedThousandthsFloat(curr_data, IMURegisters.NAVX_REG_LINEAR_ACC_X_L-first_address)
        ahrspos_update.world_linear_accel_y  = AHRSProtocol.decodeProtocolSignedThousandthsFloat(curr_data, IMURegisters.NAVX_REG_LINEAR_ACC_Y_L-first_address)
        ahrspos_update.world_linear_accel_z  = AHRSProtocol.decodeProtocolSignedThousandthsFloat(curr_data, IMURegisters.NAVX_REG_LINEAR_ACC_Z_L-first_address)
        ahrspos_update.altitude        = AHRSProtocol.decodeProtocol1616Float(curr_data, IMURegisters.NAVX_REG_ALTITUDE_D_L - first_address)
        ahrspos_update.baro_pressure   = AHRSProtocol.decodeProtocol1616Float(curr_data, IMURegisters.NAVX_REG_PRESSURE_DL - first_address)
        ahrspos_update.fused_heading   = AHRSProtocol.decodeProtocolUnsignedHundredthsFloat(curr_data, IMURegisters.NAVX_REG_FUSED_HEADING_L-first_address)
        ahrspos_update.quaternionW     = AHRSProtocol.decodeBinaryInt16(curr_data, IMURegisters.NAVX_REG_QUAT_W_L-first_address)
        ahrspos_update.quaternionX     = AHRSProtocol.decodeBinaryInt16(curr_data, IMURegisters.NAVX_REG_QUAT_X_L-first_address)
        ahrspos_update.quaternionY     = AHRSProtocol.decodeBinaryInt16(curr_data, IMURegisters.NAVX_REG_QUAT_Y_L-first_address)
        ahrspos_update.quaternionZ     = AHRSProtocol.decodeBinaryInt16(curr_data, IMURegisters.NAVX_REG_QUAT_Z_L-first_address)
        if displacement_registers:
            ahrspos_update.vel_x       = AHRSProtocol.decodeProtocol1616Float(curr_data, IMURegisters.NAVX_REG_VEL_X_I_L-first_address)
            ahrspos_update.vel_y       = AHRSProtocol.decodeProtocol1616Float(curr_data, IMURegisters.NAVX_REG_VEL_Y_I_L-first_address)
            ahrspos_update.vel_z       = AHRSProtocol.decodeProtocol1616Float(curr_data, IMURegisters.NAVX_REG_VEL_Z_I_L-first_address)
            ahrspos_update.disp_x      = AHRSProtocol.decodeProtocol1616Float(curr_data, IMURegisters.NAVX_REG_DISP_X_I_L-first_address)
            ahrspos_update.disp_y      = AHRSProtocol.decodeProtocol1616Float(curr_data, IMURegisters.NAVX_REG_DISP_Y_I_L-first_address)
            ahrspos_update.disp_z      = AHRSProtocol.decodeProtocol1616Float(curr_data, IMURegisters.NAVX_REG_DISP_Z_I_L-first_address)            
             
            self.notify_sink._setAHRSPosData(ahrspos_update)
        else:
            self.notify_sink._setAHRSData(ahrspos_update)
        
        board_state = self.board_state
        board_state.cal_status      = curr_data[IMURegisters.NAVX_REG_CAL_STATUS-first_address]
        board_state.op_status       = curr_data[IMURegisters.NAVX_REG_OP_STATUS-first_address]
        board_state.selftest_status = curr_data[IMURegisters.NAVX_REG_SELFTEST_STATUS-first_address]
        board_state.sensor_status   = AHRSProtocol.decodeBinaryUint16(curr_data,IMURegisters.NAVX_REG_SENSOR_STATUS_L-first_address)
        board_state.update_rate_hz  = curr_data[IMURegisters.NAVX_REG_UPDATE_RATE_HZ-first_address]
        board_state.gyro_fsr_dps    = AHRSProtocol.decodeBinaryUint16(curr_data,IMURegisters.NAVX_REG_GYRO_FSR_DPS_L)
        board_state.accel_fsr_g     = curr_data[IMURegisters.NAVX_REG_ACCEL_FSR_G]
        board_state.capability_flags= AHRSProtocol.decodeBinaryUint16(curr_data,IMURegisters.NAVX_REG_CAPABILITY_FLAGS_L-first_address)
        self.notify_sink._setBoardState(board_state)
        
        raw_data_update = self.raw_data_update
        raw_data_update.raw_gyro_x      = AHRSProtocol.decodeBinaryInt16(curr_data,  IMURegisters.NAVX_REG_GYRO_X_L-first_address)
        raw_data_update.raw_gyro_y      = AHRSProtocol.decodeBinaryInt16(curr_data,  IMURegisters.NAVX_REG_GYRO_Y_L-first_address)
        raw_data_update.raw_gyro_z      = AHRSProtocol.decodeBinaryInt16(curr_data,  IMURegisters.NAVX_REG_GYRO_Z_L-first_address)
        raw_data_update.raw_accel_x     = AHRSProtocol.decodeBinaryInt16(curr_data,  IMURegisters.NAVX_REG_ACC_X_L-first_address)
        raw_data_update.raw_accel_y     = AHRSProtocol.decodeBinaryInt16(curr_data,  IMURegisters.NAVX_REG_ACC_Y_L-first_address)
        raw_data_update.raw_accel_z     = AHRSProtocol.decodeBinaryInt16(curr_data,  IMURegisters.NAVX_REG_ACC_Z_L-first_address)
        raw_data_update.cal_mag_x       = AHRSProtocol.decodeBinaryInt16(curr_data,  IMURegisters.NAVX_REG_MAG_X_L-first_address)
        raw_data_update.cal_mag_y       = AHRSProtocol.decodeBinaryInt16(curr_data,  IMURegisters.NAVX_REG_MAG_Y_L-first_address)
        raw_data_update.cal_mag_z       = AHRSProtocol.decodeBinaryInt16(curr_data,  IMURegisters.NAVX_REG_MAG_Z_L-first_address)
        raw_data_update.mpu_temp_c      = ahrspos_update.mpu_temp
        self.notify_sink._setRawData(raw_data_update)
        
        self.last_update_time = Timer.getFPGATimestamp()
        self.byte_count += len(curr_data)
        self.update_count += 1
        
    def isConnected(self):
        time_since_last_update = Timer.getFPGATimestamp() - self.last_update_time
        return time_since_last_update <= IO_TIMEOUT_SECONDS
    
    def getByteCount(self):
        return self.byte_count
    
    def getUpdateCount(self):
        return self.update_count
    
    def setUpdateRateHz(self, update_rate):
        self.io_provider.write(IMURegisters.NAVX_REG_UPDATE_RATE_HZ, update_rate)
    
    def zeroYaw(self):
        self.io_provider.write( IMURegisters.NAVX_REG_INTEGRATION_CTL, 
                                   AHRSProtocol.NAVX_INTEGRATION_CTL_RESET_YAW )

    def zeroDisplacement(self):
        self.io_provider.write( IMURegisters.NAVX_REG_INTEGRATION_CTL, 
                                (AHRSProtocol.NAVX_INTEGRATION_CTL_RESET_DISP_X |
                                 AHRSProtocol.NAVX_INTEGRATION_CTL_RESET_DISP_Y |
                                 AHRSProtocol.NAVX_INTEGRATION_CTL_RESET_DISP_Z ) )       
